/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "flattr.cflattr.builder",
        "sources": [
            "src/flattr/cflattr/builder.pyx"
        ]
    },
    "module_name": "flattr.cflattr.builder"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flattr__cflattr__builder
#define __PYX_HAVE_API__flattr__cflattr__builder
/* Early includes */
#include <string.h>
#include <stdint.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/flattr/cflattr/builder.pyx",
  "stringsource",
  "src/flattr/cflattr/number_types.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6flattr_7cflattr_12number_types_Flags;
struct __pyx_obj_6flattr_7cflattr_7builder_Builder;
struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType;
struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable;
union __pyx_t_6flattr_7cflattr_7builder_double2bytes;
union __pyx_t_6flattr_7cflattr_7builder_float2bytes;
struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString;

/* "flattr/cflattr/builder.pyx":25
 * from flatbuffers import encode
 * 
 * cdef struct FlatbufferType:             # <<<<<<<<<<<<<<
 *     uint8_t bytewidth
 * 
 */
struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType {
  uint8_t bytewidth;
};

/* "flattr/cflattr/builder.pyx":29
 * 
 * 
 * cdef struct StoredVtable:             # <<<<<<<<<<<<<<
 *     uint32_t num_fields
 *     unsigned char* vtable
 */
struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable {
  uint32_t num_fields;
  unsigned char *vtable;
  int64_t offset;
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *next;
};

/* "flattr/cflattr/builder.pyx":112
 * 
 * 
 * cdef union double2bytes:             # <<<<<<<<<<<<<<
 *     double d
 *     unsigned char b[8]
 */
union __pyx_t_6flattr_7cflattr_7builder_double2bytes {
  double d;
  unsigned char b[8];
};

/* "flattr/cflattr/builder.pyx":130
 * 
 * 
 * cdef union float2bytes:             # <<<<<<<<<<<<<<
 *     float f
 *     unsigned char b[8]
 */
union __pyx_t_6flattr_7cflattr_7builder_float2bytes {
  float f;
  unsigned char b[8];
};

/* "flattr/cflattr/builder.pyx":531
 *         return self.Offset()
 * 
 *     cpdef Py_ssize_t CreateString(self, str s, encoding='utf-8', errors='strict'):             # <<<<<<<<<<<<<<
 *         """CreateString writes a null-terminated byte string as a vector."""
 *         cdef bytes x
 */
struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString {
  int __pyx_n;
  PyObject *encoding;
  PyObject *errors;
};

/* "number_types.pxd":3
 * from libc.stdint cimport uint64_t
 * 
 * cdef class Flags:             # <<<<<<<<<<<<<<
 *     cdef public unsigned short bytewidth
 *     cdef public uint64_t min_val
 */
struct __pyx_obj_6flattr_7cflattr_12number_types_Flags {
  PyObject_HEAD
  unsigned short bytewidth;
  uint64_t min_val;
  uint64_t max_val;
  PyObject *py_type;
  PyObject *name;
  PyObject *packer_type;
  int should_check;
};


/* "flattr/cflattr/builder.pyx":195
 * cdef unsigned int MAX_BUFFER_SIZE = 2**31
 * 
 * cdef class Builder(object):             # <<<<<<<<<<<<<<
 *     """ A Builder is used to construct one or more FlatBuffers.
 * 
 */
struct __pyx_obj_6flattr_7cflattr_7builder_Builder {
  PyObject_HEAD
  struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *__pyx_vtab;
  unsigned char *buffer;
  Py_ssize_t buffer_length;
  Py_ssize_t *current_vtable;
  Py_ssize_t current_vtable_length;
  Py_ssize_t head;
  Py_ssize_t minalign;
  int objectEnd;
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *vtables;
  int nested;
  int finished;
  Py_ssize_t vectorNumElems;
};



struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder {
  Py_ssize_t (*WriteVtable)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *);
  void (*growByteBuffer)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *);
  Py_ssize_t (*Offset)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *);
  PyObject *(*Pad)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t);
  void (*Prep)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, Py_ssize_t);
  void (*PrependSOffsetTRelative)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int);
  void (*PrependUOffsetTRelative)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, int __pyx_skip_dispatch);
  Py_ssize_t (*EndVector)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int __pyx_skip_dispatch);
  Py_ssize_t (*CreateString)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString *__pyx_optional_args);
  Py_ssize_t (*CreateByteVector)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, int __pyx_skip_dispatch);
  void (*assertNested)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *);
  void (*assertNotNested)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *);
  void (*assertStructIsInline)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t);
  void (*Slot)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t);
  Py_ssize_t (*__pyx___Finish)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, int);
  void (*Prepend)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *, int64_t);
  void (*PrependSlot)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *, Py_ssize_t, int64_t, int64_t);
  PyObject *(*PrependStructSlot)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, PyObject *, PyObject *);
  PyObject *(*PrependInt32)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int32_t, int __pyx_skip_dispatch);
  void (*PrependVOffsetT)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint16_t);
  void (*Place)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint64_t, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *);
  void (*PlaceVOffsetT)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint16_t);
  void (*PlaceSOffsetT)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int32_t);
  void (*PlaceUOffsetT)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint32_t);
};
static struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *__pyx_vtabptr_6flattr_7cflattr_7builder_Builder;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_WriteVtable(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_growByteBuffer(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_Offset(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_Pad(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_n); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Prep(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_size, Py_ssize_t __pyx_v_additionalBytes); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_off); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependUOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_off, int __pyx_skip_dispatch); /* proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_EndVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_CreateString(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch, struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString *__pyx_optional_args); /* proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_CreateByteVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertNested(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertNotNested(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertStructIsInline(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_obj); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_slotnum); /* proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder___Finish(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_rootTable, int __pyx_v_sizePrefix); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Prepend(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_type, int64_t __pyx_v_off); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_type, Py_ssize_t __pyx_v_o, int64_t __pyx_v_x, int64_t __pyx_v_d); /* proto*/
static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependStructSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_v, PyObject *__pyx_v_x, PyObject *__pyx_v_d); /* proto*/
static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependInt32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependVOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint16_t __pyx_v_x); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Place(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint64_t __pyx_v_x, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_fb_type); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceVOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint16_t __pyx_v_x); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceSOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x); /* proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceUOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint32_t __pyx_v_x); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'flattr.cflattr.number_types' */
static PyTypeObject *__pyx_ptype_6flattr_7cflattr_12number_types_Flags = 0;
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_12number_types_enforce_number(uint64_t, struct __pyx_obj_6flattr_7cflattr_12number_types_Flags *); /*proto*/

/* Module declarations from 'flattr.cflattr.builder' */
static PyTypeObject *__pyx_ptype_6flattr_7cflattr_7builder_Builder = 0;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_bool_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_int8_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_int16_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_int64_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t;
static struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t;
static unsigned int __pyx_v_6flattr_7cflattr_7builder_VtableMetadataFields;
static unsigned int __pyx_v_6flattr_7cflattr_7builder_MAX_BUFFER_SIZE;
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeBool(int, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint8(uint8_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint16(uint16_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint32(uint32_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint64(uint64_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt8(int8_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt16(int16_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt32(int32_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt64(int64_t, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeFloat64(double, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeFloat32(float, unsigned char *, Py_ssize_t, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "flattr.cflattr.builder"
extern int __pyx_module_is_main_flattr__cflattr__builder;
int __pyx_module_is_main_flattr__cflattr__builder = 0;

/* Implementation of 'flattr.cflattr.builder' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_d[] = "d";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_head[] = "head";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_strict[] = "strict";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_Builder[] = "Builder";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_elemSize[] = "elemSize";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_numElems[] = "numElems";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_EndVector[] = "EndVector";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_alignment[] = "alignment";
static const char __pyx_k_bytewidth[] = "bytewidth";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pack_into[] = "pack_into";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_flatbuffers[] = "flatbuffers";
static const char __pyx_k_initialSize[] = "initialSize";
static const char __pyx_k_CreateString[] = "CreateString";
static const char __pyx_k_PrependInt32[] = "PrependInt32";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_IsNestedError[] = "IsNestedError";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_BuilderSizeError[] = "BuilderSizeError";
static const char __pyx_k_CreateByteVector[] = "CreateByteVector";
static const char __pyx_k_IsNotNestedError[] = "IsNotNestedError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_OffsetArithmeticError[] = "OffsetArithmeticError";
static const char __pyx_k_StructIsNotInlineError[] = "StructIsNotInlineError";
static const char __pyx_k_flattr_cflattr_builder[] = "flattr.cflattr.builder";
static const char __pyx_k_BuilderNotFinishedError[] = "BuilderNotFinishedError";
static const char __pyx_k_PrependUOffsetTRelative[] = "PrependUOffsetTRelative";
static const char __pyx_k_Error_caused_by_an_Offset_arith[] = "\n    Error caused by an Offset arithmetic error. Probably caused by bad\n    writing of fields. This is considered an unreachable situation in\n    normal circumstances.\n    ";
static const char __pyx_k_Error_caused_by_causing_a_Build[] = "\n    Error caused by causing a Builder to exceed the hardcoded limit of 2\n    gigabytes.\n    ";
static const char __pyx_k_Error_caused_by_not_calling_Fin[] = "\n    Error caused by not calling `Finish` before calling `Output`.\n    ";
static const char __pyx_k_Error_caused_by_using_a_Builder[] = "\n    Error caused by using a Builder to write Object data when not inside\n    an Object.\n    ";
static const char __pyx_k_flatbuffers_Cannot_create_Builde[] = "flatbuffers: Cannot create Builder larger than 2 gigabytes.";
static const char __pyx_k_flatbuffers_Offset_arithmetic_er[] = "flatbuffers: Offset arithmetic error.";
static const char __pyx_k_flatbuffers_Tried_to_write_a_Str[] = "flatbuffers: Tried to write a Struct at an Offset that is different from the current Offset of the Builder.";
static const char __pyx_k_flatbuffers_cannot_grow_buffer_b[] = "flatbuffers: cannot grow buffer beyond 2 gigabytes";
static const char __pyx_k_self_current_vtable_self_vtables[] = "self.current_vtable,self.vtables cannot be converted to a Python object for pickling";
static const char __pyx_k_Error_caused_by_using_a_Builder_2[] = "\n    Error caused by using a Builder to begin an Object when an Object is\n    already being built.\n    ";
static const char __pyx_k_Error_caused_by_using_a_Builder_3[] = "\n    Error caused by using a Builder to write a Struct at a location that\n    is not the current Offset.\n    ";
static PyObject *__pyx_n_s_Builder;
static PyObject *__pyx_n_s_BuilderNotFinishedError;
static PyObject *__pyx_n_s_BuilderSizeError;
static PyObject *__pyx_n_s_CreateByteVector;
static PyObject *__pyx_n_s_CreateString;
static PyObject *__pyx_n_s_EndVector;
static PyObject *__pyx_kp_s_Error_caused_by_an_Offset_arith;
static PyObject *__pyx_kp_s_Error_caused_by_causing_a_Build;
static PyObject *__pyx_kp_s_Error_caused_by_not_calling_Fin;
static PyObject *__pyx_kp_s_Error_caused_by_using_a_Builder;
static PyObject *__pyx_kp_s_Error_caused_by_using_a_Builder_2;
static PyObject *__pyx_kp_s_Error_caused_by_using_a_Builder_3;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IsNestedError;
static PyObject *__pyx_n_s_IsNotNestedError;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_OffsetArithmeticError;
static PyObject *__pyx_n_s_PrependInt32;
static PyObject *__pyx_n_s_PrependUOffsetTRelative;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_StructIsNotInlineError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_alignment;
static PyObject *__pyx_n_s_buffer;
static PyObject *__pyx_n_s_bytewidth;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_elemSize;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_n_s_errors;
static PyObject *__pyx_n_s_flatbuffers;
static PyObject *__pyx_kp_u_flatbuffers_Cannot_create_Builde;
static PyObject *__pyx_kp_u_flatbuffers_Offset_arithmetic_er;
static PyObject *__pyx_kp_u_flatbuffers_Tried_to_write_a_Str;
static PyObject *__pyx_kp_u_flatbuffers_cannot_grow_buffer_b;
static PyObject *__pyx_n_s_flattr_cflattr_builder;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_head;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_initialSize;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_numElems;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_n_s_pack_into;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_kp_s_self_current_vtable_self_vtables;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_u_strict;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_writeBool(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_2writeUint8(CYTHON_UNUSED PyObject *__pyx_self, uint8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_4writeUint16(CYTHON_UNUSED PyObject *__pyx_self, uint16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_6writeUint32(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_8writeUint64(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_10writeInt8(CYTHON_UNUSED PyObject *__pyx_self, int8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_12writeInt16(CYTHON_UNUSED PyObject *__pyx_self, int16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_14writeInt32(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_16writeInt64(CYTHON_UNUSED PyObject *__pyx_self, int64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_18writeFloat64(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_20writeFloat32(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head); /* proto */
static int __pyx_pf_6flattr_7cflattr_7builder_7Builder___init__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_initialSize); /* proto */
static void __pyx_pf_6flattr_7cflattr_7builder_7Builder_2__dealloc__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_4Output(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_5Bytes___get__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_6StartObject(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_numfields); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_8EndObject(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_10PrependUOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_off); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_12StartVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_elemSize, Py_ssize_t __pyx_v_numElems, Py_ssize_t __pyx_v_alignment); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_14EndVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_16CreateString(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_18CreateByteVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_20Finish(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_rootTable); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_22FinishSizePrefixed(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_rootTable); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_24PrependBoolSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int __pyx_v_x, int __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_26PrependByteSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int __pyx_v_x, int __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_28PrependUint8Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint8_t __pyx_v_x, uint8_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_30PrependUint16Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint16_t __pyx_v_x, uint16_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_32PrependUint32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint32_t __pyx_v_x, uint32_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_34PrependUint64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint64_t __pyx_v_x, uint64_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_36PrependInt8Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int8_t __pyx_v_x, int8_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_38PrependInt16Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int16_t __pyx_v_x, int16_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_40PrependInt32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int32_t __pyx_v_x, int32_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_42PrependInt64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int64_t __pyx_v_x, int64_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_44PrependFloat32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, float __pyx_v_x, float __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_46PrependFloat64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, double __pyx_v_x, double __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_48PrependUOffsetTRelativeSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, Py_ssize_t __pyx_v_x, Py_ssize_t __pyx_v_d); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_50PrependBool(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_52PrependByte(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_54PrependUint8(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_56PrependUint16(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_58PrependUint32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_60PrependUint64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_62PrependInt8(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_64PrependInt16(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_66PrependInt32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_68PrependInt64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_70PrependFloat32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_72PrependFloat64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_74__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_76__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6flattr_7cflattr_7builder_Builder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
/* Late includes */

/* "flattr/cflattr/builder.pyx":49
 * 
 * 
 * cdef inline void Write(packer_type, buf, Py_ssize_t head, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *     """ Write encodes `n` at buf[head] using `packer_type`. """
 *     packer_type.pack_into(buf, head, n)
 */

static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_Write(PyObject *__pyx_v_packer_type, PyObject *__pyx_v_buf, Py_ssize_t __pyx_v_head, Py_ssize_t __pyx_v_n) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Write", 0);

  /* "flattr/cflattr/builder.pyx":51
 * cdef inline void Write(packer_type, buf, Py_ssize_t head, Py_ssize_t n):
 *     """ Write encodes `n` at buf[head] using `packer_type`. """
 *     packer_type.pack_into(buf, head, n)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_packer_type, __pyx_n_s_pack_into); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_head); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_buf, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_buf, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_buf);
    __Pyx_GIVEREF(__pyx_v_buf);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_buf);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flattr/cflattr/builder.pyx":49
 * 
 * 
 * cdef inline void Write(packer_type, buf, Py_ssize_t head, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *     """ Write encodes `n` at buf[head] using `packer_type`. """
 *     packer_type.pack_into(buf, head, n)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":54
 * 
 * 
 * cpdef inline void writeBool(bint n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = 1 if n else 0
 * 
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_1writeBool(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeBool(int __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("writeBool", 0);

  /* "flattr/cflattr/builder.pyx":55
 * 
 * cpdef inline void writeBool(bint n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head] = 1 if n else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_n != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  (__pyx_v_buffer[__pyx_v_head]) = __pyx_t_1;

  /* "flattr/cflattr/builder.pyx":54
 * 
 * 
 * cpdef inline void writeBool(bint n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = 1 if n else 0
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_1writeBool(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_1writeBool(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeBool (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeBool", 1, 3, 3, 1); __PYX_ERR(0, 54, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeBool", 1, 3, 3, 2); __PYX_ERR(0, 54, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeBool") < 0)) __PYX_ERR(0, 54, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeBool", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 54, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeBool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_writeBool(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_writeBool(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeBool", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeBool(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeBool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":58
 * 
 * 
 * cpdef inline void writeUint8(uint8_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_3writeUint8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint8(uint8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint8", 0);

  /* "flattr/cflattr/builder.pyx":59
 * 
 * cpdef inline void writeUint8(uint8_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":58
 * 
 * 
 * cpdef inline void writeUint8(uint8_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_3writeUint8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_3writeUint8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint8_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint8 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint8", 1, 3, 3, 1); __PYX_ERR(0, 58, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint8", 1, 3, 3, 2); __PYX_ERR(0, 58, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeUint8") < 0)) __PYX_ERR(0, 58, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_uint8_t(values[0]); if (unlikely((__pyx_v_n == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeUint8", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 58, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_2writeUint8(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_2writeUint8(CYTHON_UNUSED PyObject *__pyx_self, uint8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeUint8", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeUint8(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":62
 * 
 * 
 * cpdef inline void writeUint16(uint16_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_5writeUint16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint16(uint16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint16", 0);

  /* "flattr/cflattr/builder.pyx":63
 * 
 * cpdef inline void writeUint16(uint16_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":64
 * cpdef inline void writeUint16(uint16_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":62
 * 
 * 
 * cpdef inline void writeUint16(uint16_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_5writeUint16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_5writeUint16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint16_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint16 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint16", 1, 3, 3, 1); __PYX_ERR(0, 62, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint16", 1, 3, 3, 2); __PYX_ERR(0, 62, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeUint16") < 0)) __PYX_ERR(0, 62, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_uint16_t(values[0]); if (unlikely((__pyx_v_n == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeUint16", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 62, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_4writeUint16(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_4writeUint16(CYTHON_UNUSED PyObject *__pyx_self, uint16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeUint16", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeUint16(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":67
 * 
 * 
 * cpdef inline void writeUint32(uint32_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7writeUint32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint32(uint32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint32", 0);

  /* "flattr/cflattr/builder.pyx":68
 * 
 * cpdef inline void writeUint32(uint32_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+3] = (n >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = ((__pyx_v_n >> 24) & 0xFF);

  /* "flattr/cflattr/builder.pyx":69
 * cpdef inline void writeUint32(uint32_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = ((__pyx_v_n >> 16) & 0xFF);

  /* "flattr/cflattr/builder.pyx":70
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":71
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":67
 * 
 * 
 * cpdef inline void writeUint32(uint32_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7writeUint32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7writeUint32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint32 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint32", 1, 3, 3, 1); __PYX_ERR(0, 67, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint32", 1, 3, 3, 2); __PYX_ERR(0, 67, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeUint32") < 0)) __PYX_ERR(0, 67, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_n == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeUint32", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_6writeUint32(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_6writeUint32(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeUint32", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeUint32(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":74
 * 
 * 
 * cpdef inline void writeUint64(uint64_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_9writeUint64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeUint64(uint64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint64", 0);

  /* "flattr/cflattr/builder.pyx":75
 * 
 * cpdef inline void writeUint64(uint64_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+7] = (n >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 7)]) = ((__pyx_v_n >> 56) & 0xFF);

  /* "flattr/cflattr/builder.pyx":76
 * cpdef inline void writeUint64(uint64_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 6)]) = ((__pyx_v_n >> 48) & 0xFF);

  /* "flattr/cflattr/builder.pyx":77
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 5)]) = ((__pyx_v_n >> 40) & 0xFF);

  /* "flattr/cflattr/builder.pyx":78
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 4)]) = ((__pyx_v_n >> 32) & 0xFF);

  /* "flattr/cflattr/builder.pyx":79
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = ((__pyx_v_n >> 24) & 0xFF);

  /* "flattr/cflattr/builder.pyx":80
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = ((__pyx_v_n >> 16) & 0xFF);

  /* "flattr/cflattr/builder.pyx":81
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":82
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":74
 * 
 * 
 * cpdef inline void writeUint64(uint64_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_9writeUint64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_9writeUint64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint64_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeUint64 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint64", 1, 3, 3, 1); __PYX_ERR(0, 74, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeUint64", 1, 3, 3, 2); __PYX_ERR(0, 74, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeUint64") < 0)) __PYX_ERR(0, 74, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_uint64_t(values[0]); if (unlikely((__pyx_v_n == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 74, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 74, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 74, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeUint64", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 74, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_8writeUint64(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_8writeUint64(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeUint64", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeUint64(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeUint64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":85
 * 
 * 
 * cpdef inline void writeInt8(int8_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_11writeInt8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt8(int8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt8", 0);

  /* "flattr/cflattr/builder.pyx":86
 * 
 * cpdef inline void writeInt8(int8_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":85
 * 
 * 
 * cpdef inline void writeInt8(int8_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_11writeInt8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_11writeInt8(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int8_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt8 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt8", 1, 3, 3, 1); __PYX_ERR(0, 85, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt8", 1, 3, 3, 2); __PYX_ERR(0, 85, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeInt8") < 0)) __PYX_ERR(0, 85, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_int8_t(values[0]); if (unlikely((__pyx_v_n == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeInt8", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 85, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_10writeInt8(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_10writeInt8(CYTHON_UNUSED PyObject *__pyx_self, int8_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeInt8", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeInt8(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":89
 * 
 * 
 * cpdef inline void writeInt16(int16_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_13writeInt16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt16(int16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt16", 0);

  /* "flattr/cflattr/builder.pyx":90
 * 
 * cpdef inline void writeInt16(int16_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":91
 * cpdef inline void writeInt16(int16_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":89
 * 
 * 
 * cpdef inline void writeInt16(int16_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_13writeInt16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_13writeInt16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int16_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt16 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt16", 1, 3, 3, 1); __PYX_ERR(0, 89, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt16", 1, 3, 3, 2); __PYX_ERR(0, 89, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeInt16") < 0)) __PYX_ERR(0, 89, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_int16_t(values[0]); if (unlikely((__pyx_v_n == ((int16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeInt16", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 89, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_12writeInt16(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_12writeInt16(CYTHON_UNUSED PyObject *__pyx_self, int16_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeInt16", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeInt16(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":94
 * 
 * 
 * cpdef inline void writeInt32(int32_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_15writeInt32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt32(int32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt32", 0);

  /* "flattr/cflattr/builder.pyx":95
 * 
 * cpdef inline void writeInt32(int32_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+3] = (n >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = ((__pyx_v_n >> 24) & 0xFF);

  /* "flattr/cflattr/builder.pyx":96
 * cpdef inline void writeInt32(int32_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = ((__pyx_v_n >> 16) & 0xFF);

  /* "flattr/cflattr/builder.pyx":97
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":98
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":94
 * 
 * 
 * cpdef inline void writeInt32(int32_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_15writeInt32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_15writeInt32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int32_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt32 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt32", 1, 3, 3, 1); __PYX_ERR(0, 94, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt32", 1, 3, 3, 2); __PYX_ERR(0, 94, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeInt32") < 0)) __PYX_ERR(0, 94, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_int32_t(values[0]); if (unlikely((__pyx_v_n == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 94, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 94, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 94, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeInt32", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 94, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_14writeInt32(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_14writeInt32(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeInt32", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeInt32(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":101
 * 
 * 
 * cpdef inline void writeInt64(int64_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_17writeInt64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeInt64(int64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt64", 0);

  /* "flattr/cflattr/builder.pyx":102
 * 
 * cpdef inline void writeInt64(int64_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+7] = (n >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 7)]) = ((__pyx_v_n >> 56) & 0xFF);

  /* "flattr/cflattr/builder.pyx":103
 * cpdef inline void writeInt64(int64_t n, unsigned char* buffer, Py_ssize_t head):
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 6)]) = ((__pyx_v_n >> 48) & 0xFF);

  /* "flattr/cflattr/builder.pyx":104
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 5)]) = ((__pyx_v_n >> 40) & 0xFF);

  /* "flattr/cflattr/builder.pyx":105
 *     buffer[head+6] = (n >> 48) & 0xFF
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 4)]) = ((__pyx_v_n >> 32) & 0xFF);

  /* "flattr/cflattr/builder.pyx":106
 *     buffer[head+5] = (n >> 40) & 0xFF
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = ((__pyx_v_n >> 24) & 0xFF);

  /* "flattr/cflattr/builder.pyx":107
 *     buffer[head+4] = (n >> 32) & 0xFF
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = ((__pyx_v_n >> 16) & 0xFF);

  /* "flattr/cflattr/builder.pyx":108
 *     buffer[head+3] = (n >> 24) & 0xFF
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[head] = n & 0xFF
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = ((__pyx_v_n >> 8) & 0xFF);

  /* "flattr/cflattr/builder.pyx":109
 *     buffer[head+2] = (n >> 16) & 0xFF
 *     buffer[head+1] = (n >> 8) & 0xFF
 *     buffer[head] = n & 0xFF             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_n & 0xFF);

  /* "flattr/cflattr/builder.pyx":101
 * 
 * 
 * cpdef inline void writeInt64(int64_t n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     buffer[head+7] = (n >> 56) & 0xFF
 *     buffer[head+6] = (n >> 48) & 0xFF
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_17writeInt64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_17writeInt64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int64_t __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeInt64 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt64", 1, 3, 3, 1); __PYX_ERR(0, 101, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeInt64", 1, 3, 3, 2); __PYX_ERR(0, 101, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeInt64") < 0)) __PYX_ERR(0, 101, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_n == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeInt64", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 101, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_16writeInt64(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_16writeInt64(CYTHON_UNUSED PyObject *__pyx_self, int64_t __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeInt64", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeInt64(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeInt64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":117
 * 
 * 
 * cpdef inline void writeFloat64(double n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     cdef double2bytes t
 *     t.d = n
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_19writeFloat64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeFloat64(double __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  union __pyx_t_6flattr_7cflattr_7builder_double2bytes __pyx_v_t;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeFloat64", 0);

  /* "flattr/cflattr/builder.pyx":119
 * cpdef inline void writeFloat64(double n, unsigned char* buffer, Py_ssize_t head):
 *     cdef double2bytes t
 *     t.d = n             # <<<<<<<<<<<<<<
 *     buffer[head+7] = t.b[7]
 *     buffer[head+6] = t.b[6]
 */
  __pyx_v_t.d = __pyx_v_n;

  /* "flattr/cflattr/builder.pyx":120
 *     cdef double2bytes t
 *     t.d = n
 *     buffer[head+7] = t.b[7]             # <<<<<<<<<<<<<<
 *     buffer[head+6] = t.b[6]
 *     buffer[head+5] = t.b[5]
 */
  (__pyx_v_buffer[(__pyx_v_head + 7)]) = (__pyx_v_t.b[7]);

  /* "flattr/cflattr/builder.pyx":121
 *     t.d = n
 *     buffer[head+7] = t.b[7]
 *     buffer[head+6] = t.b[6]             # <<<<<<<<<<<<<<
 *     buffer[head+5] = t.b[5]
 *     buffer[head+4] = t.b[4]
 */
  (__pyx_v_buffer[(__pyx_v_head + 6)]) = (__pyx_v_t.b[6]);

  /* "flattr/cflattr/builder.pyx":122
 *     buffer[head+7] = t.b[7]
 *     buffer[head+6] = t.b[6]
 *     buffer[head+5] = t.b[5]             # <<<<<<<<<<<<<<
 *     buffer[head+4] = t.b[4]
 *     buffer[head+3] = t.b[3]
 */
  (__pyx_v_buffer[(__pyx_v_head + 5)]) = (__pyx_v_t.b[5]);

  /* "flattr/cflattr/builder.pyx":123
 *     buffer[head+6] = t.b[6]
 *     buffer[head+5] = t.b[5]
 *     buffer[head+4] = t.b[4]             # <<<<<<<<<<<<<<
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]
 */
  (__pyx_v_buffer[(__pyx_v_head + 4)]) = (__pyx_v_t.b[4]);

  /* "flattr/cflattr/builder.pyx":124
 *     buffer[head+5] = t.b[5]
 *     buffer[head+4] = t.b[4]
 *     buffer[head+3] = t.b[3]             # <<<<<<<<<<<<<<
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = (__pyx_v_t.b[3]);

  /* "flattr/cflattr/builder.pyx":125
 *     buffer[head+4] = t.b[4]
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]             # <<<<<<<<<<<<<<
 *     buffer[head+1] = t.b[1]
 *     buffer[head] = t.b[0]
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = (__pyx_v_t.b[2]);

  /* "flattr/cflattr/builder.pyx":126
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]             # <<<<<<<<<<<<<<
 *     buffer[head] = t.b[0]
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = (__pyx_v_t.b[1]);

  /* "flattr/cflattr/builder.pyx":127
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]
 *     buffer[head] = t.b[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_t.b[0]);

  /* "flattr/cflattr/builder.pyx":117
 * 
 * 
 * cpdef inline void writeFloat64(double n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     cdef double2bytes t
 *     t.d = n
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_19writeFloat64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_19writeFloat64(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeFloat64 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeFloat64", 1, 3, 3, 1); __PYX_ERR(0, 117, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeFloat64", 1, 3, 3, 2); __PYX_ERR(0, 117, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeFloat64") < 0)) __PYX_ERR(0, 117, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_n == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeFloat64", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeFloat64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_18writeFloat64(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_18writeFloat64(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeFloat64", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeFloat64(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeFloat64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":135
 * 
 * 
 * cpdef inline void writeFloat32(float n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     cdef float2bytes t
 *     t.f = n
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_21writeFloat32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_7builder_writeFloat32(float __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head, CYTHON_UNUSED int __pyx_skip_dispatch) {
  union __pyx_t_6flattr_7cflattr_7builder_float2bytes __pyx_v_t;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeFloat32", 0);

  /* "flattr/cflattr/builder.pyx":137
 * cpdef inline void writeFloat32(float n, unsigned char* buffer, Py_ssize_t head):
 *     cdef float2bytes t
 *     t.f = n             # <<<<<<<<<<<<<<
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]
 */
  __pyx_v_t.f = __pyx_v_n;

  /* "flattr/cflattr/builder.pyx":138
 *     cdef float2bytes t
 *     t.f = n
 *     buffer[head+3] = t.b[3]             # <<<<<<<<<<<<<<
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]
 */
  (__pyx_v_buffer[(__pyx_v_head + 3)]) = (__pyx_v_t.b[3]);

  /* "flattr/cflattr/builder.pyx":139
 *     t.f = n
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]             # <<<<<<<<<<<<<<
 *     buffer[head+1] = t.b[1]
 *     buffer[head] = t.b[0]
 */
  (__pyx_v_buffer[(__pyx_v_head + 2)]) = (__pyx_v_t.b[2]);

  /* "flattr/cflattr/builder.pyx":140
 *     buffer[head+3] = t.b[3]
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]             # <<<<<<<<<<<<<<
 *     buffer[head] = t.b[0]
 * 
 */
  (__pyx_v_buffer[(__pyx_v_head + 1)]) = (__pyx_v_t.b[1]);

  /* "flattr/cflattr/builder.pyx":141
 *     buffer[head+2] = t.b[2]
 *     buffer[head+1] = t.b[1]
 *     buffer[head] = t.b[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[__pyx_v_head]) = (__pyx_v_t.b[0]);

  /* "flattr/cflattr/builder.pyx":135
 * 
 * 
 * cpdef inline void writeFloat32(float n, unsigned char* buffer, Py_ssize_t head):             # <<<<<<<<<<<<<<
 *     cdef float2bytes t
 *     t.f = n
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_21writeFloat32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_21writeFloat32(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  float __pyx_v_n;
  unsigned char *__pyx_v_buffer;
  Py_ssize_t __pyx_v_head;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writeFloat32 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,&__pyx_n_s_buffer,&__pyx_n_s_head,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeFloat32", 1, 3, 3, 1); __PYX_ERR(0, 135, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_head)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writeFloat32", 1, 3, 3, 2); __PYX_ERR(0, 135, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writeFloat32") < 0)) __PYX_ERR(0, 135, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_n = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_n == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 135, __pyx_L3_error)
    __pyx_v_buffer = __Pyx_PyObject_AsWritableUString(values[1]); if (unlikely((!__pyx_v_buffer) && PyErr_Occurred())) __PYX_ERR(0, 135, __pyx_L3_error)
    __pyx_v_head = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_head == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 135, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writeFloat32", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 135, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.writeFloat32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_20writeFloat32(__pyx_self, __pyx_v_n, __pyx_v_buffer, __pyx_v_head);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_20writeFloat32(CYTHON_UNUSED PyObject *__pyx_self, float __pyx_v_n, unsigned char *__pyx_v_buffer, Py_ssize_t __pyx_v_head) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeFloat32", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_writeFloat32(__pyx_v_n, __pyx_v_buffer, __pyx_v_head, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.writeFloat32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":227
 *     cdef Py_ssize_t vectorNumElems
 * 
 *     def __init__(self, Py_ssize_t initialSize):             # <<<<<<<<<<<<<<
 *         """Initializes a Builder of size `initial_size`.
 * 
 */

/* Python wrapper */
static int __pyx_pw_6flattr_7cflattr_7builder_7Builder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder___init__[] = "Initializes a Builder of size `initial_size`.\n\n        The internal buffer is grown as needed.\n        ";
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_6flattr_7cflattr_7builder_7Builder___init__;
#endif
static int __pyx_pw_6flattr_7cflattr_7builder_7Builder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_initialSize;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_initialSize,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_initialSize)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 227, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_initialSize = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_initialSize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 227, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder___init__(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_initialSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6flattr_7cflattr_7builder_7Builder___init__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_initialSize) {
  PyObject *__pyx_v_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "flattr/cflattr/builder.pyx":233
 *         """
 * 
 *         if not (0 <= initialSize <= MAX_BUFFER_SIZE):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."
 *             raise BuilderSizeError(msg)
 */
  __pyx_t_1 = (0 <= __pyx_v_initialSize);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_initialSize <= __pyx_v_6flattr_7cflattr_7builder_MAX_BUFFER_SIZE);
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "flattr/cflattr/builder.pyx":234
 * 
 *         if not (0 <= initialSize <= MAX_BUFFER_SIZE):
 *             msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."             # <<<<<<<<<<<<<<
 *             raise BuilderSizeError(msg)
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
 */
    __Pyx_INCREF(__pyx_kp_u_flatbuffers_Cannot_create_Builde);
    __pyx_v_msg = __pyx_kp_u_flatbuffers_Cannot_create_Builde;

    /* "flattr/cflattr/builder.pyx":235
 *         if not (0 <= initialSize <= MAX_BUFFER_SIZE):
 *             msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."
 *             raise BuilderSizeError(msg)             # <<<<<<<<<<<<<<
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
 *         if self.buffer is NULL:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BuilderSizeError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 235, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":233
 *         """
 * 
 *         if not (0 <= initialSize <= MAX_BUFFER_SIZE):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."
 *             raise BuilderSizeError(msg)
 */
  }

  /* "flattr/cflattr/builder.pyx":236
 *             msg = "flatbuffers: Cannot create Builder larger than 2 gigabytes."
 *             raise BuilderSizeError(msg)
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)             # <<<<<<<<<<<<<<
 *         if self.buffer is NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->buffer = ((unsigned char *)PyMem_Malloc(((sizeof(unsigned char)) * __pyx_v_initialSize)));

  /* "flattr/cflattr/builder.pyx":237
 *             raise BuilderSizeError(msg)
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
 *         if self.buffer is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
 */
  __pyx_t_2 = ((__pyx_v_self->buffer == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "flattr/cflattr/builder.pyx":238
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
 *         if self.buffer is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
 *         self.buffer_length = initialSize
 */
    PyErr_NoMemory(); __PYX_ERR(0, 238, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":237
 *             raise BuilderSizeError(msg)
 *         self.buffer = <unsigned char*> PyMem_Malloc(sizeof(unsigned char) * initialSize)
 *         if self.buffer is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
 */
  }

  /* "flattr/cflattr/builder.pyx":239
 *         if self.buffer is NULL:
 *             raise MemoryError()
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)             # <<<<<<<<<<<<<<
 *         self.buffer_length = initialSize
 *         self.current_vtable = NULL
 */
  (void)(memset(__pyx_v_self->buffer, 0, ((sizeof(unsigned char)) * __pyx_v_initialSize)));

  /* "flattr/cflattr/builder.pyx":240
 *             raise MemoryError()
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
 *         self.buffer_length = initialSize             # <<<<<<<<<<<<<<
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0
 */
  __pyx_v_self->buffer_length = __pyx_v_initialSize;

  /* "flattr/cflattr/builder.pyx":241
 *         memset(self.buffer, 0, sizeof(unsigned char) * initialSize)
 *         self.buffer_length = initialSize
 *         self.current_vtable = NULL             # <<<<<<<<<<<<<<
 *         self.current_vtable_length = 0
 *         self.head = initialSize
 */
  __pyx_v_self->current_vtable = NULL;

  /* "flattr/cflattr/builder.pyx":242
 *         self.buffer_length = initialSize
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0             # <<<<<<<<<<<<<<
 *         self.head = initialSize
 *         self.minalign = 1
 */
  __pyx_v_self->current_vtable_length = 0;

  /* "flattr/cflattr/builder.pyx":243
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0
 *         self.head = initialSize             # <<<<<<<<<<<<<<
 *         self.minalign = 1
 *         self.objectEnd = -1
 */
  __pyx_v_self->head = __pyx_v_initialSize;

  /* "flattr/cflattr/builder.pyx":244
 *         self.current_vtable_length = 0
 *         self.head = initialSize
 *         self.minalign = 1             # <<<<<<<<<<<<<<
 *         self.objectEnd = -1
 *         self.nested = False
 */
  __pyx_v_self->minalign = 1;

  /* "flattr/cflattr/builder.pyx":245
 *         self.head = initialSize
 *         self.minalign = 1
 *         self.objectEnd = -1             # <<<<<<<<<<<<<<
 *         self.nested = False
 *         self.finished = False
 */
  __pyx_v_self->objectEnd = -1;

  /* "flattr/cflattr/builder.pyx":246
 *         self.minalign = 1
 *         self.objectEnd = -1
 *         self.nested = False             # <<<<<<<<<<<<<<
 *         self.finished = False
 *         self.vtables = NULL
 */
  __pyx_v_self->nested = 0;

  /* "flattr/cflattr/builder.pyx":247
 *         self.objectEnd = -1
 *         self.nested = False
 *         self.finished = False             # <<<<<<<<<<<<<<
 *         self.vtables = NULL
 *         self.vectorNumElems = 0
 */
  __pyx_v_self->finished = 0;

  /* "flattr/cflattr/builder.pyx":248
 *         self.nested = False
 *         self.finished = False
 *         self.vtables = NULL             # <<<<<<<<<<<<<<
 *         self.vectorNumElems = 0
 * 
 */
  __pyx_v_self->vtables = NULL;

  /* "flattr/cflattr/builder.pyx":249
 *         self.finished = False
 *         self.vtables = NULL
 *         self.vectorNumElems = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->vectorNumElems = 0;

  /* "flattr/cflattr/builder.pyx":227
 *     cdef Py_ssize_t vectorNumElems
 * 
 *     def __init__(self, Py_ssize_t initialSize):             # <<<<<<<<<<<<<<
 *         """Initializes a Builder of size `initial_size`.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":251
 *         self.vectorNumElems = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.buffer)
 *         if self.current_vtable != NULL:
 */

/* Python wrapper */
static void __pyx_pw_6flattr_7cflattr_7builder_7Builder_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6flattr_7cflattr_7builder_7Builder_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6flattr_7cflattr_7builder_7Builder_2__dealloc__(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6flattr_7cflattr_7builder_7Builder_2__dealloc__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *__pyx_v_next;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *__pyx_t_2;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "flattr/cflattr/builder.pyx":252
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self.buffer)             # <<<<<<<<<<<<<<
 *         if self.current_vtable != NULL:
 *             PyMem_Free(self.current_vtable)
 */
  PyMem_Free(__pyx_v_self->buffer);

  /* "flattr/cflattr/builder.pyx":253
 *     def __dealloc__(self):
 *         PyMem_Free(self.buffer)
 *         if self.current_vtable != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->current_vtable != NULL) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":254
 *         PyMem_Free(self.buffer)
 *         if self.current_vtable != NULL:
 *             PyMem_Free(self.current_vtable)             # <<<<<<<<<<<<<<
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0
 */
    PyMem_Free(__pyx_v_self->current_vtable);

    /* "flattr/cflattr/builder.pyx":255
 *         if self.current_vtable != NULL:
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL             # <<<<<<<<<<<<<<
 *             self.current_vtable_length = 0
 *         while self.vtables is not NULL:
 */
    __pyx_v_self->current_vtable = NULL;

    /* "flattr/cflattr/builder.pyx":256
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0             # <<<<<<<<<<<<<<
 *         while self.vtables is not NULL:
 *             next = self.vtables.next
 */
    __pyx_v_self->current_vtable_length = 0;

    /* "flattr/cflattr/builder.pyx":253
 *     def __dealloc__(self):
 *         PyMem_Free(self.buffer)
 *         if self.current_vtable != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 */
  }

  /* "flattr/cflattr/builder.pyx":257
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0
 *         while self.vtables is not NULL:             # <<<<<<<<<<<<<<
 *             next = self.vtables.next
 *             PyMem_Free(self.vtables.vtable)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_self->vtables != NULL) != 0);
    if (!__pyx_t_1) break;

    /* "flattr/cflattr/builder.pyx":258
 *             self.current_vtable_length = 0
 *         while self.vtables is not NULL:
 *             next = self.vtables.next             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.vtables.vtable)
 *             PyMem_Free(self.vtables)
 */
    __pyx_t_2 = __pyx_v_self->vtables->next;
    __pyx_v_next = __pyx_t_2;

    /* "flattr/cflattr/builder.pyx":259
 *         while self.vtables is not NULL:
 *             next = self.vtables.next
 *             PyMem_Free(self.vtables.vtable)             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.vtables)
 *             self.vtables = next
 */
    PyMem_Free(__pyx_v_self->vtables->vtable);

    /* "flattr/cflattr/builder.pyx":260
 *             next = self.vtables.next
 *             PyMem_Free(self.vtables.vtable)
 *             PyMem_Free(self.vtables)             # <<<<<<<<<<<<<<
 *             self.vtables = next
 * 
 */
    PyMem_Free(__pyx_v_self->vtables);

    /* "flattr/cflattr/builder.pyx":261
 *             PyMem_Free(self.vtables.vtable)
 *             PyMem_Free(self.vtables)
 *             self.vtables = next             # <<<<<<<<<<<<<<
 * 
 *     def Output(self):
 */
    __pyx_v_self->vtables = __pyx_v_next;
  }

  /* "flattr/cflattr/builder.pyx":251
 *         self.vectorNumElems = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.buffer)
 *         if self.current_vtable != NULL:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":263
 *             self.vtables = next
 * 
 *     def Output(self):             # <<<<<<<<<<<<<<
 *         """Return the portion of the buffer that has been used for writing data.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_5Output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_4Output[] = "Return the portion of the buffer that has been used for writing data.\n\n        This is the typical way to access the FlatBuffer data inside the\n        builder. If you try to access `Builder.Bytes` directly, you would need\n        to manually index it with `Head()`, since the buffer is constructed\n        backwards.\n\n        It raises BuilderNotFinishedError if the buffer has not been finished\n        with `Finish`.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_5Output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Output (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_4Output(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_4Output(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  PyObject *__pyx_v_res = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Output", 0);

  /* "flattr/cflattr/builder.pyx":275
 *         """
 * 
 *         if not self.finished:             # <<<<<<<<<<<<<<
 *             raise BuilderNotFinishedError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->finished != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":276
 * 
 *         if not self.finished:
 *             raise BuilderNotFinishedError()             # <<<<<<<<<<<<<<
 * 
 *         cdef bytes res = self.buffer[self.head:self.buffer_length]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_BuilderNotFinishedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 276, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":275
 *         """
 * 
 *         if not self.finished:             # <<<<<<<<<<<<<<
 *             raise BuilderNotFinishedError()
 * 
 */
  }

  /* "flattr/cflattr/builder.pyx":278
 *             raise BuilderNotFinishedError()
 * 
 *         cdef bytes res = self.buffer[self.head:self.buffer_length]             # <<<<<<<<<<<<<<
 * 
 *         return res
 */
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_self->buffer) + __pyx_v_self->head, __pyx_v_self->buffer_length - __pyx_v_self->head); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_res = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":280
 *         cdef bytes res = self.buffer[self.head:self.buffer_length]
 * 
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":263
 *             self.vtables = next
 * 
 *     def Output(self):             # <<<<<<<<<<<<<<
 *         """Return the portion of the buffer that has been used for writing data.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.Output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":283
 * 
 *     @property
 *     def Bytes(self):             # <<<<<<<<<<<<<<
 *         return self.buffer[0:self.buffer_length]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_5Bytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_5Bytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_5Bytes___get__(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_5Bytes___get__(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "flattr/cflattr/builder.pyx":284
 *     @property
 *     def Bytes(self):
 *         return self.buffer[0:self.buffer_length]             # <<<<<<<<<<<<<<
 * 
 *     def StartObject(self, int numfields):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_self->buffer) + 0, __pyx_v_self->buffer_length - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":283
 * 
 *     @property
 *     def Bytes(self):             # <<<<<<<<<<<<<<
 *         return self.buffer[0:self.buffer_length]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.Bytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":286
 *         return self.buffer[0:self.buffer_length]
 * 
 *     def StartObject(self, int numfields):             # <<<<<<<<<<<<<<
 *         """StartObject initializes bookkeeping for writing a new object."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_7StartObject(PyObject *__pyx_v_self, PyObject *__pyx_arg_numfields); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_6StartObject[] = "StartObject initializes bookkeeping for writing a new object.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_7StartObject(PyObject *__pyx_v_self, PyObject *__pyx_arg_numfields) {
  int __pyx_v_numfields;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("StartObject (wrapper)", 0);
  assert(__pyx_arg_numfields); {
    __pyx_v_numfields = __Pyx_PyInt_As_int(__pyx_arg_numfields); if (unlikely((__pyx_v_numfields == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.StartObject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_6StartObject(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((int)__pyx_v_numfields));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_6StartObject(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_numfields) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("StartObject", 0);

  /* "flattr/cflattr/builder.pyx":289
 *         """StartObject initializes bookkeeping for writing a new object."""
 * 
 *         self.assertNotNested()             # <<<<<<<<<<<<<<
 * 
 *         if numfields > 0:
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNotNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":291
 *         self.assertNotNested()
 * 
 *         if numfields > 0:             # <<<<<<<<<<<<<<
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
 *             if self.current_vtable is NULL:
 */
  __pyx_t_1 = ((__pyx_v_numfields > 0) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":292
 * 
 *         if numfields > 0:
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)             # <<<<<<<<<<<<<<
 *             if self.current_vtable is NULL:
 *                 raise MemoryError()
 */
    __pyx_v_self->current_vtable = ((Py_ssize_t *)PyMem_Malloc(((sizeof(Py_ssize_t)) * __pyx_v_numfields)));

    /* "flattr/cflattr/builder.pyx":293
 *         if numfields > 0:
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
 *             if self.current_vtable is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
 */
    __pyx_t_1 = ((__pyx_v_self->current_vtable == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "flattr/cflattr/builder.pyx":294
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
 *             if self.current_vtable is NULL:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
 *         self.current_vtable_length = numfields
 */
      PyErr_NoMemory(); __PYX_ERR(0, 294, __pyx_L1_error)

      /* "flattr/cflattr/builder.pyx":293
 *         if numfields > 0:
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
 *             if self.current_vtable is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
 */
    }

    /* "flattr/cflattr/builder.pyx":295
 *             if self.current_vtable is NULL:
 *                 raise MemoryError()
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)             # <<<<<<<<<<<<<<
 *         self.current_vtable_length = numfields
 *         self.objectEnd = self.Offset()
 */
    (void)(memset(__pyx_v_self->current_vtable, 0, ((sizeof(Py_ssize_t)) * __pyx_v_numfields)));

    /* "flattr/cflattr/builder.pyx":291
 *         self.assertNotNested()
 * 
 *         if numfields > 0:             # <<<<<<<<<<<<<<
 *             self.current_vtable = <Py_ssize_t*>PyMem_Malloc(sizeof(Py_ssize_t) * numfields)
 *             if self.current_vtable is NULL:
 */
  }

  /* "flattr/cflattr/builder.pyx":296
 *                 raise MemoryError()
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
 *         self.current_vtable_length = numfields             # <<<<<<<<<<<<<<
 *         self.objectEnd = self.Offset()
 *         self.nested = True
 */
  __pyx_v_self->current_vtable_length = __pyx_v_numfields;

  /* "flattr/cflattr/builder.pyx":297
 *             memset(self.current_vtable, 0, sizeof(Py_ssize_t) * numfields)
 *         self.current_vtable_length = numfields
 *         self.objectEnd = self.Offset()             # <<<<<<<<<<<<<<
 *         self.nested = True
 * 
 */
  __pyx_v_self->objectEnd = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);

  /* "flattr/cflattr/builder.pyx":298
 *         self.current_vtable_length = numfields
 *         self.objectEnd = self.Offset()
 *         self.nested = True             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t WriteVtable(self):
 */
  __pyx_v_self->nested = 1;

  /* "flattr/cflattr/builder.pyx":286
 *         return self.buffer[0:self.buffer_length]
 * 
 *     def StartObject(self, int numfields):             # <<<<<<<<<<<<<<
 *         """StartObject initializes bookkeeping for writing a new object."""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.StartObject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":300
 *         self.nested = True
 * 
 *     cdef Py_ssize_t WriteVtable(self):             # <<<<<<<<<<<<<<
 *         """
 *         WriteVtable serializes the vtable for the current object, if needed.
 */

static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_WriteVtable(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  Py_ssize_t __pyx_v_objectOffset;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_off;
  Py_ssize_t __pyx_v_objectSize;
  Py_ssize_t __pyx_v_objectStart;
  Py_ssize_t __pyx_v_vBytes;
  Py_ssize_t __pyx_v_effective_vtable_length;
  Py_ssize_t __pyx_v_vtable_offset;
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *__pyx_v_svt;
  Py_ssize_t __pyx_v_norm_vtable_length;
  unsigned char *__pyx_v_norm_vtable;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *__pyx_t_3;
  int __pyx_t_4;
  int64_t __pyx_t_5;
  __Pyx_RefNannySetupContext("WriteVtable", 0);

  /* "flattr/cflattr/builder.pyx":328
 *         # Prepend a zero scalar to the object. Later in this function we'll
 *         # write an offset here that points to the object's vtable:
 *         self.PrependSOffsetTRelative(0)  # 32-bit value.             # <<<<<<<<<<<<<<
 * 
 *         objectOffset = self.Offset()
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSOffsetTRelative(__pyx_v_self, 0);

  /* "flattr/cflattr/builder.pyx":330
 *         self.PrependSOffsetTRelative(0)  # 32-bit value.
 * 
 *         objectOffset = self.Offset()             # <<<<<<<<<<<<<<
 * 
 *         # Trim trailing 0 offsets.
 */
  __pyx_v_objectOffset = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);

  /* "flattr/cflattr/builder.pyx":333
 * 
 *         # Trim trailing 0 offsets.
 *         effective_vtable_length = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.current_vtable_length, 0, -1):
 *             if self.current_vtable[i-1] != 0:
 */
  __pyx_v_effective_vtable_length = 0;

  /* "flattr/cflattr/builder.pyx":334
 *         # Trim trailing 0 offsets.
 *         effective_vtable_length = 0
 *         for i in range(self.current_vtable_length, 0, -1):             # <<<<<<<<<<<<<<
 *             if self.current_vtable[i-1] != 0:
 *                 effective_vtable_length = i
 */
  for (__pyx_t_1 = __pyx_v_self->current_vtable_length; __pyx_t_1 > 0; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "flattr/cflattr/builder.pyx":335
 *         effective_vtable_length = 0
 *         for i in range(self.current_vtable_length, 0, -1):
 *             if self.current_vtable[i-1] != 0:             # <<<<<<<<<<<<<<
 *                 effective_vtable_length = i
 *                 break
 */
    __pyx_t_2 = (((__pyx_v_self->current_vtable[(__pyx_v_i - 1)]) != 0) != 0);
    if (__pyx_t_2) {

      /* "flattr/cflattr/builder.pyx":336
 *         for i in range(self.current_vtable_length, 0, -1):
 *             if self.current_vtable[i-1] != 0:
 *                 effective_vtable_length = i             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_effective_vtable_length = __pyx_v_i;

      /* "flattr/cflattr/builder.pyx":337
 *             if self.current_vtable[i-1] != 0:
 *                 effective_vtable_length = i
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         # Write out the current vtable in reverse, because
 */
      goto __pyx_L4_break;

      /* "flattr/cflattr/builder.pyx":335
 *         effective_vtable_length = 0
 *         for i in range(self.current_vtable_length, 0, -1):
 *             if self.current_vtable[i-1] != 0:             # <<<<<<<<<<<<<<
 *                 effective_vtable_length = i
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "flattr/cflattr/builder.pyx":341
 *         # Write out the current vtable in reverse, because
 *         # serialization occurs in last-first order:
 *         i = effective_vtable_length - 1             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t norm_vtable_length = sizeof(uint16_t) * effective_vtable_length
 *         norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)
 */
  __pyx_v_i = (__pyx_v_effective_vtable_length - 1);

  /* "flattr/cflattr/builder.pyx":342
 *         # serialization occurs in last-first order:
 *         i = effective_vtable_length - 1
 *         cdef Py_ssize_t norm_vtable_length = sizeof(uint16_t) * effective_vtable_length             # <<<<<<<<<<<<<<
 *         norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)
 *         while i >= 0:
 */
  __pyx_v_norm_vtable_length = ((sizeof(uint16_t)) * __pyx_v_effective_vtable_length);

  /* "flattr/cflattr/builder.pyx":343
 *         i = effective_vtable_length - 1
 *         cdef Py_ssize_t norm_vtable_length = sizeof(uint16_t) * effective_vtable_length
 *         norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)             # <<<<<<<<<<<<<<
 *         while i >= 0:
 *             off = 0
 */
  __pyx_v_norm_vtable = ((unsigned char *)PyMem_Malloc(__pyx_v_norm_vtable_length));

  /* "flattr/cflattr/builder.pyx":344
 *         cdef Py_ssize_t norm_vtable_length = sizeof(uint16_t) * effective_vtable_length
 *         norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)
 *         while i >= 0:             # <<<<<<<<<<<<<<
 *             off = 0
 *             if self.current_vtable[i] != 0:
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i >= 0) != 0);
    if (!__pyx_t_2) break;

    /* "flattr/cflattr/builder.pyx":345
 *         norm_vtable = <unsigned char*>PyMem_Malloc(norm_vtable_length)
 *         while i >= 0:
 *             off = 0             # <<<<<<<<<<<<<<
 *             if self.current_vtable[i] != 0:
 *                 # Forward reference to field;
 */
    __pyx_v_off = 0;

    /* "flattr/cflattr/builder.pyx":346
 *         while i >= 0:
 *             off = 0
 *             if self.current_vtable[i] != 0:             # <<<<<<<<<<<<<<
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 */
    __pyx_t_2 = (((__pyx_v_self->current_vtable[__pyx_v_i]) != 0) != 0);
    if (__pyx_t_2) {

      /* "flattr/cflattr/builder.pyx":349
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 *                 off = objectOffset - self.current_vtable[i]             # <<<<<<<<<<<<<<
 * 
 *             writeUint16(off, norm_vtable, i*2)
 */
      __pyx_v_off = (__pyx_v_objectOffset - (__pyx_v_self->current_vtable[__pyx_v_i]));

      /* "flattr/cflattr/builder.pyx":346
 *         while i >= 0:
 *             off = 0
 *             if self.current_vtable[i] != 0:             # <<<<<<<<<<<<<<
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 */
    }

    /* "flattr/cflattr/builder.pyx":351
 *                 off = objectOffset - self.current_vtable[i]
 * 
 *             writeUint16(off, norm_vtable, i*2)             # <<<<<<<<<<<<<<
 *             i -= 1
 * 
 */
    __pyx_f_6flattr_7cflattr_7builder_writeUint16(__pyx_v_off, __pyx_v_norm_vtable, (__pyx_v_i * 2), 0);

    /* "flattr/cflattr/builder.pyx":352
 * 
 *             writeUint16(off, norm_vtable, i*2)
 *             i -= 1             # <<<<<<<<<<<<<<
 * 
 *         # Check if this vtable has been seen already.
 */
    __pyx_v_i = (__pyx_v_i - 1);
  }

  /* "flattr/cflattr/builder.pyx":355
 * 
 *         # Check if this vtable has been seen already.
 *         svt = self.vtables             # <<<<<<<<<<<<<<
 *         while svt is not NULL:
 *             if (svt.num_fields == effective_vtable_length and
 */
  __pyx_t_3 = __pyx_v_self->vtables;
  __pyx_v_svt = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":356
 *         # Check if this vtable has been seen already.
 *         svt = self.vtables
 *         while svt is not NULL:             # <<<<<<<<<<<<<<
 *             if (svt.num_fields == effective_vtable_length and
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_svt != NULL) != 0);
    if (!__pyx_t_2) break;

    /* "flattr/cflattr/builder.pyx":357
 *         svt = self.vtables
 *         while svt is not NULL:
 *             if (svt.num_fields == effective_vtable_length and             # <<<<<<<<<<<<<<
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 *                 break
 */
    __pyx_t_4 = ((__pyx_v_svt->num_fields == __pyx_v_effective_vtable_length) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L12_bool_binop_done;
    }

    /* "flattr/cflattr/builder.pyx":358
 *         while svt is not NULL:
 *             if (svt.num_fields == effective_vtable_length and
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):             # <<<<<<<<<<<<<<
 *                 break
 *             svt = svt.next
 */
    __pyx_t_4 = ((memcmp(__pyx_v_norm_vtable, __pyx_v_svt->vtable, __pyx_v_norm_vtable_length) == 0) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L12_bool_binop_done:;

    /* "flattr/cflattr/builder.pyx":357
 *         svt = self.vtables
 *         while svt is not NULL:
 *             if (svt.num_fields == effective_vtable_length and             # <<<<<<<<<<<<<<
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 *                 break
 */
    if (__pyx_t_2) {

      /* "flattr/cflattr/builder.pyx":359
 *             if (svt.num_fields == effective_vtable_length and
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 *                 break             # <<<<<<<<<<<<<<
 *             svt = svt.next
 * 
 */
      goto __pyx_L10_break;

      /* "flattr/cflattr/builder.pyx":357
 *         svt = self.vtables
 *         while svt is not NULL:
 *             if (svt.num_fields == effective_vtable_length and             # <<<<<<<<<<<<<<
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 *                 break
 */
    }

    /* "flattr/cflattr/builder.pyx":360
 *                 memcmp(norm_vtable, svt.vtable, norm_vtable_length) == 0):
 *                 break
 *             svt = svt.next             # <<<<<<<<<<<<<<
 * 
 *         if svt is not NULL:
 */
    __pyx_t_3 = __pyx_v_svt->next;
    __pyx_v_svt = __pyx_t_3;
  }
  __pyx_L10_break:;

  /* "flattr/cflattr/builder.pyx":362
 *             svt = svt.next
 * 
 *         if svt is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 */
  __pyx_t_2 = ((__pyx_v_svt != NULL) != 0);
  if (__pyx_t_2) {

    /* "flattr/cflattr/builder.pyx":363
 * 
 *         if svt is not NULL:
 *             PyMem_Free(self.current_vtable)             # <<<<<<<<<<<<<<
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0
 */
    PyMem_Free(__pyx_v_self->current_vtable);

    /* "flattr/cflattr/builder.pyx":364
 *         if svt is not NULL:
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL             # <<<<<<<<<<<<<<
 *             self.current_vtable_length = 0
 *             PyMem_Free(norm_vtable)
 */
    __pyx_v_self->current_vtable = NULL;

    /* "flattr/cflattr/builder.pyx":365
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0             # <<<<<<<<<<<<<<
 *             PyMem_Free(norm_vtable)
 *             objectStart = self.buffer_length - objectOffset
 */
    __pyx_v_self->current_vtable_length = 0;

    /* "flattr/cflattr/builder.pyx":366
 *             self.current_vtable = NULL
 *             self.current_vtable_length = 0
 *             PyMem_Free(norm_vtable)             # <<<<<<<<<<<<<<
 *             objectStart = self.buffer_length - objectOffset
 *             self.head = objectStart
 */
    PyMem_Free(__pyx_v_norm_vtable);

    /* "flattr/cflattr/builder.pyx":367
 *             self.current_vtable_length = 0
 *             PyMem_Free(norm_vtable)
 *             objectStart = self.buffer_length - objectOffset             # <<<<<<<<<<<<<<
 *             self.head = objectStart
 * 
 */
    __pyx_v_objectStart = (__pyx_v_self->buffer_length - __pyx_v_objectOffset);

    /* "flattr/cflattr/builder.pyx":368
 *             PyMem_Free(norm_vtable)
 *             objectStart = self.buffer_length - objectOffset
 *             self.head = objectStart             # <<<<<<<<<<<<<<
 * 
 *             # Write the offset to the found vtable in the
 */
    __pyx_v_self->head = __pyx_v_objectStart;

    /* "flattr/cflattr/builder.pyx":373
 *             # already-allocated SOffsetT at the beginning of this object.
 *             # The cached vtable contains its offset from the end of the buffer.
 *             vtable_offset = svt.offset             # <<<<<<<<<<<<<<
 *             writeInt32(vtable_offset - objectOffset, self.buffer, objectStart)
 *             return self.Offset()
 */
    __pyx_t_5 = __pyx_v_svt->offset;
    __pyx_v_vtable_offset = __pyx_t_5;

    /* "flattr/cflattr/builder.pyx":374
 *             # The cached vtable contains its offset from the end of the buffer.
 *             vtable_offset = svt.offset
 *             writeInt32(vtable_offset - objectOffset, self.buffer, objectStart)             # <<<<<<<<<<<<<<
 *             return self.Offset()
 * 
 */
    __pyx_f_6flattr_7cflattr_7builder_writeInt32((__pyx_v_vtable_offset - __pyx_v_objectOffset), __pyx_v_self->buffer, __pyx_v_objectStart, 0);

    /* "flattr/cflattr/builder.pyx":375
 *             vtable_offset = svt.offset
 *             writeInt32(vtable_offset - objectOffset, self.buffer, objectStart)
 *             return self.Offset()             # <<<<<<<<<<<<<<
 * 
 *         i = effective_vtable_length - 1
 */
    __pyx_r = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);
    goto __pyx_L0;

    /* "flattr/cflattr/builder.pyx":362
 *             svt = svt.next
 * 
 *         if svt is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.current_vtable)
 *             self.current_vtable = NULL
 */
  }

  /* "flattr/cflattr/builder.pyx":377
 *             return self.Offset()
 * 
 *         i = effective_vtable_length - 1             # <<<<<<<<<<<<<<
 *         while i >= 0:
 *             off = 0
 */
  __pyx_v_i = (__pyx_v_effective_vtable_length - 1);

  /* "flattr/cflattr/builder.pyx":378
 * 
 *         i = effective_vtable_length - 1
 *         while i >= 0:             # <<<<<<<<<<<<<<
 *             off = 0
 *             if self.current_vtable[i] != 0:
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i >= 0) != 0);
    if (!__pyx_t_2) break;

    /* "flattr/cflattr/builder.pyx":379
 *         i = effective_vtable_length - 1
 *         while i >= 0:
 *             off = 0             # <<<<<<<<<<<<<<
 *             if self.current_vtable[i] != 0:
 *                 # Forward reference to field;
 */
    __pyx_v_off = 0;

    /* "flattr/cflattr/builder.pyx":380
 *         while i >= 0:
 *             off = 0
 *             if self.current_vtable[i] != 0:             # <<<<<<<<<<<<<<
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 */
    __pyx_t_2 = (((__pyx_v_self->current_vtable[__pyx_v_i]) != 0) != 0);
    if (__pyx_t_2) {

      /* "flattr/cflattr/builder.pyx":383
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 *                 off = objectOffset - self.current_vtable[i]             # <<<<<<<<<<<<<<
 * 
 *             self.PrependVOffsetT(off)
 */
      __pyx_v_off = (__pyx_v_objectOffset - (__pyx_v_self->current_vtable[__pyx_v_i]));

      /* "flattr/cflattr/builder.pyx":380
 *         while i >= 0:
 *             off = 0
 *             if self.current_vtable[i] != 0:             # <<<<<<<<<<<<<<
 *                 # Forward reference to field;
 *                 # use 32bit number to ensure no overflow:
 */
    }

    /* "flattr/cflattr/builder.pyx":385
 *                 off = objectOffset - self.current_vtable[i]
 * 
 *             self.PrependVOffsetT(off)             # <<<<<<<<<<<<<<
 *             i -= 1
 *         # The two metadata fields are written last.
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependVOffsetT(__pyx_v_self, __pyx_v_off);

    /* "flattr/cflattr/builder.pyx":386
 * 
 *             self.PrependVOffsetT(off)
 *             i -= 1             # <<<<<<<<<<<<<<
 *         # The two metadata fields are written last.
 * 
 */
    __pyx_v_i = (__pyx_v_i - 1);
  }

  /* "flattr/cflattr/builder.pyx":390
 * 
 *         # First, store the object bytesize:
 *         objectSize = objectOffset - self.objectEnd             # <<<<<<<<<<<<<<
 *         self.PrependVOffsetT(objectSize)
 * 
 */
  __pyx_v_objectSize = (__pyx_v_objectOffset - __pyx_v_self->objectEnd);

  /* "flattr/cflattr/builder.pyx":391
 *         # First, store the object bytesize:
 *         objectSize = objectOffset - self.objectEnd
 *         self.PrependVOffsetT(objectSize)             # <<<<<<<<<<<<<<
 * 
 *         # Second, store the vtable bytesize:
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependVOffsetT(__pyx_v_self, __pyx_v_objectSize);

  /* "flattr/cflattr/builder.pyx":394
 * 
 *         # Second, store the vtable bytesize:
 *         vBytes = effective_vtable_length + VtableMetadataFields             # <<<<<<<<<<<<<<
 *         vBytes *= Fb_uint16_t.bytewidth
 *         self.PrependVOffsetT(vBytes)
 */
  __pyx_v_vBytes = (__pyx_v_effective_vtable_length + __pyx_v_6flattr_7cflattr_7builder_VtableMetadataFields);

  /* "flattr/cflattr/builder.pyx":395
 *         # Second, store the vtable bytesize:
 *         vBytes = effective_vtable_length + VtableMetadataFields
 *         vBytes *= Fb_uint16_t.bytewidth             # <<<<<<<<<<<<<<
 *         self.PrependVOffsetT(vBytes)
 * 
 */
  __pyx_v_vBytes = (__pyx_v_vBytes * __pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":396
 *         vBytes = effective_vtable_length + VtableMetadataFields
 *         vBytes *= Fb_uint16_t.bytewidth
 *         self.PrependVOffsetT(vBytes)             # <<<<<<<<<<<<<<
 * 
 *         # Next, write the offset to the new vtable in the
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependVOffsetT(__pyx_v_self, __pyx_v_vBytes);

  /* "flattr/cflattr/builder.pyx":400
 *         # Next, write the offset to the new vtable in the
 *         # already-allocated SOffsetT at the beginning of this object:
 *         objectStart = self.buffer_length - objectOffset             # <<<<<<<<<<<<<<
 *         writeUint32(self.Offset() - objectOffset, self.buffer, objectStart)
 * 
 */
  __pyx_v_objectStart = (__pyx_v_self->buffer_length - __pyx_v_objectOffset);

  /* "flattr/cflattr/builder.pyx":401
 *         # already-allocated SOffsetT at the beginning of this object:
 *         objectStart = self.buffer_length - objectOffset
 *         writeUint32(self.Offset() - objectOffset, self.buffer, objectStart)             # <<<<<<<<<<<<<<
 * 
 *         # Finally, store this vtable in memory for future
 */
  __pyx_f_6flattr_7cflattr_7builder_writeUint32((((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self) - __pyx_v_objectOffset), __pyx_v_self->buffer, __pyx_v_objectStart, 0);

  /* "flattr/cflattr/builder.pyx":405
 *         # Finally, store this vtable in memory for future
 *         # deduplication:
 *         svt = <StoredVtable*>PyMem_Malloc(sizeof(StoredVtable))             # <<<<<<<<<<<<<<
 *         svt.num_fields = effective_vtable_length
 *         svt.vtable = norm_vtable
 */
  __pyx_v_svt = ((struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable *)PyMem_Malloc((sizeof(struct __pyx_t_6flattr_7cflattr_7builder_StoredVtable))));

  /* "flattr/cflattr/builder.pyx":406
 *         # deduplication:
 *         svt = <StoredVtable*>PyMem_Malloc(sizeof(StoredVtable))
 *         svt.num_fields = effective_vtable_length             # <<<<<<<<<<<<<<
 *         svt.vtable = norm_vtable
 *         svt.next = self.vtables
 */
  __pyx_v_svt->num_fields = __pyx_v_effective_vtable_length;

  /* "flattr/cflattr/builder.pyx":407
 *         svt = <StoredVtable*>PyMem_Malloc(sizeof(StoredVtable))
 *         svt.num_fields = effective_vtable_length
 *         svt.vtable = norm_vtable             # <<<<<<<<<<<<<<
 *         svt.next = self.vtables
 *         svt.offset = self.Offset()
 */
  __pyx_v_svt->vtable = __pyx_v_norm_vtable;

  /* "flattr/cflattr/builder.pyx":408
 *         svt.num_fields = effective_vtable_length
 *         svt.vtable = norm_vtable
 *         svt.next = self.vtables             # <<<<<<<<<<<<<<
 *         svt.offset = self.Offset()
 *         self.vtables = svt
 */
  __pyx_t_3 = __pyx_v_self->vtables;
  __pyx_v_svt->next = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":409
 *         svt.vtable = norm_vtable
 *         svt.next = self.vtables
 *         svt.offset = self.Offset()             # <<<<<<<<<<<<<<
 *         self.vtables = svt
 * 
 */
  __pyx_v_svt->offset = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);

  /* "flattr/cflattr/builder.pyx":410
 *         svt.next = self.vtables
 *         svt.offset = self.Offset()
 *         self.vtables = svt             # <<<<<<<<<<<<<<
 * 
 *         PyMem_Free(self.current_vtable)
 */
  __pyx_v_self->vtables = __pyx_v_svt;

  /* "flattr/cflattr/builder.pyx":412
 *         self.vtables = svt
 * 
 *         PyMem_Free(self.current_vtable)             # <<<<<<<<<<<<<<
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0
 */
  PyMem_Free(__pyx_v_self->current_vtable);

  /* "flattr/cflattr/builder.pyx":413
 * 
 *         PyMem_Free(self.current_vtable)
 *         self.current_vtable = NULL             # <<<<<<<<<<<<<<
 *         self.current_vtable_length = 0
 *         return objectOffset
 */
  __pyx_v_self->current_vtable = NULL;

  /* "flattr/cflattr/builder.pyx":414
 *         PyMem_Free(self.current_vtable)
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0             # <<<<<<<<<<<<<<
 *         return objectOffset
 * 
 */
  __pyx_v_self->current_vtable_length = 0;

  /* "flattr/cflattr/builder.pyx":415
 *         self.current_vtable = NULL
 *         self.current_vtable_length = 0
 *         return objectOffset             # <<<<<<<<<<<<<<
 * 
 *     def EndObject(self):
 */
  __pyx_r = __pyx_v_objectOffset;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":300
 *         self.nested = True
 * 
 *     cdef Py_ssize_t WriteVtable(self):             # <<<<<<<<<<<<<<
 *         """
 *         WriteVtable serializes the vtable for the current object, if needed.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":417
 *         return objectOffset
 * 
 *     def EndObject(self):             # <<<<<<<<<<<<<<
 *         """EndObject writes data necessary to finish object construction."""
 *         self.assertNested()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_9EndObject(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_8EndObject[] = "EndObject writes data necessary to finish object construction.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_9EndObject(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EndObject (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_8EndObject(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_8EndObject(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EndObject", 0);

  /* "flattr/cflattr/builder.pyx":419
 *     def EndObject(self):
 *         """EndObject writes data necessary to finish object construction."""
 *         self.assertNested()             # <<<<<<<<<<<<<<
 *         self.nested = False
 *         return self.WriteVtable()
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 419, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":420
 *         """EndObject writes data necessary to finish object construction."""
 *         self.assertNested()
 *         self.nested = False             # <<<<<<<<<<<<<<
 *         return self.WriteVtable()
 * 
 */
  __pyx_v_self->nested = 0;

  /* "flattr/cflattr/builder.pyx":421
 *         self.assertNested()
 *         self.nested = False
 *         return self.WriteVtable()             # <<<<<<<<<<<<<<
 * 
 *     cdef void growByteBuffer(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->WriteVtable(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":417
 *         return objectOffset
 * 
 *     def EndObject(self):             # <<<<<<<<<<<<<<
 *         """EndObject writes data necessary to finish object construction."""
 *         self.assertNested()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.EndObject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":423
 *         return self.WriteVtable()
 * 
 *     cdef void growByteBuffer(self):             # <<<<<<<<<<<<<<
 *         """Doubles the size of the byteslice, and copies the old data towards
 *            the end of the new buffer (since we build the buffer backwards)."""
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_growByteBuffer(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  Py_ssize_t __pyx_v_newSize;
  PyObject *__pyx_v_msg = NULL;
  unsigned char *__pyx_v_buffer_2;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("growByteBuffer", 0);

  /* "flattr/cflattr/builder.pyx":427
 *            the end of the new buffer (since we build the buffer backwards)."""
 *         cdef Py_ssize_t newSize
 *         if self.buffer_length == MAX_BUFFER_SIZE:             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: cannot grow buffer beyond 2 gigabytes"
 *             raise BuilderSizeError(msg)
 */
  __pyx_t_1 = ((__pyx_v_self->buffer_length == __pyx_v_6flattr_7cflattr_7builder_MAX_BUFFER_SIZE) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":428
 *         cdef Py_ssize_t newSize
 *         if self.buffer_length == MAX_BUFFER_SIZE:
 *             msg = "flatbuffers: cannot grow buffer beyond 2 gigabytes"             # <<<<<<<<<<<<<<
 *             raise BuilderSizeError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_u_flatbuffers_cannot_grow_buffer_b);
    __pyx_v_msg = __pyx_kp_u_flatbuffers_cannot_grow_buffer_b;

    /* "flattr/cflattr/builder.pyx":429
 *         if self.buffer_length == MAX_BUFFER_SIZE:
 *             msg = "flatbuffers: cannot grow buffer beyond 2 gigabytes"
 *             raise BuilderSizeError(msg)             # <<<<<<<<<<<<<<
 * 
 *         newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_BuilderSizeError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 429, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":427
 *            the end of the new buffer (since we build the buffer backwards)."""
 *         cdef Py_ssize_t newSize
 *         if self.buffer_length == MAX_BUFFER_SIZE:             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: cannot grow buffer beyond 2 gigabytes"
 *             raise BuilderSizeError(msg)
 */
  }

  /* "flattr/cflattr/builder.pyx":431
 *             raise BuilderSizeError(msg)
 * 
 *         newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)             # <<<<<<<<<<<<<<
 *         if newSize == 0:
 *             newSize = 1
 */
  __pyx_t_5 = __pyx_v_6flattr_7cflattr_7builder_MAX_BUFFER_SIZE;
  __pyx_t_6 = (__pyx_v_self->buffer_length * 2);
  if (((__pyx_t_5 < __pyx_t_6) != 0)) {
    __pyx_t_7 = __pyx_t_5;
  } else {
    __pyx_t_7 = __pyx_t_6;
  }
  __pyx_v_newSize = __pyx_t_7;

  /* "flattr/cflattr/builder.pyx":432
 * 
 *         newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)
 *         if newSize == 0:             # <<<<<<<<<<<<<<
 *             newSize = 1
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
 */
  __pyx_t_1 = ((__pyx_v_newSize == 0) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":433
 *         newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)
 *         if newSize == 0:
 *             newSize = 1             # <<<<<<<<<<<<<<
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
 *         memset(buffer_2, 0, sizeof(unsigned char) * newSize)
 */
    __pyx_v_newSize = 1;

    /* "flattr/cflattr/builder.pyx":432
 * 
 *         newSize = min(self.buffer_length * 2, MAX_BUFFER_SIZE)
 *         if newSize == 0:             # <<<<<<<<<<<<<<
 *             newSize = 1
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
 */
  }

  /* "flattr/cflattr/builder.pyx":434
 *         if newSize == 0:
 *             newSize = 1
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)             # <<<<<<<<<<<<<<
 *         memset(buffer_2, 0, sizeof(unsigned char) * newSize)
 *         memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)
 */
  __pyx_v_buffer_2 = ((unsigned char *)PyMem_Malloc(((sizeof(unsigned char)) * __pyx_v_newSize)));

  /* "flattr/cflattr/builder.pyx":435
 *             newSize = 1
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
 *         memset(buffer_2, 0, sizeof(unsigned char) * newSize)             # <<<<<<<<<<<<<<
 *         memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)
 *         self.buffer_length = newSize
 */
  (void)(memset(__pyx_v_buffer_2, 0, ((sizeof(unsigned char)) * __pyx_v_newSize)));

  /* "flattr/cflattr/builder.pyx":436
 *         buffer_2 = <unsigned char *>PyMem_Malloc(sizeof(unsigned char) * newSize)
 *         memset(buffer_2, 0, sizeof(unsigned char) * newSize)
 *         memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)             # <<<<<<<<<<<<<<
 *         self.buffer_length = newSize
 *         PyMem_Free(self.buffer)
 */
  (void)(memcpy((&(__pyx_v_buffer_2[(__pyx_v_newSize - __pyx_v_self->buffer_length)])), __pyx_v_self->buffer, __pyx_v_self->buffer_length));

  /* "flattr/cflattr/builder.pyx":437
 *         memset(buffer_2, 0, sizeof(unsigned char) * newSize)
 *         memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)
 *         self.buffer_length = newSize             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.buffer)
 *         self.buffer = buffer_2
 */
  __pyx_v_self->buffer_length = __pyx_v_newSize;

  /* "flattr/cflattr/builder.pyx":438
 *         memcpy(&buffer_2[newSize-self.buffer_length], self.buffer, self.buffer_length)
 *         self.buffer_length = newSize
 *         PyMem_Free(self.buffer)             # <<<<<<<<<<<<<<
 *         self.buffer = buffer_2
 * 
 */
  PyMem_Free(__pyx_v_self->buffer);

  /* "flattr/cflattr/builder.pyx":439
 *         self.buffer_length = newSize
 *         PyMem_Free(self.buffer)
 *         self.buffer = buffer_2             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t Offset(self):
 */
  __pyx_v_self->buffer = __pyx_v_buffer_2;

  /* "flattr/cflattr/builder.pyx":423
 *         return self.WriteVtable()
 * 
 *     cdef void growByteBuffer(self):             # <<<<<<<<<<<<<<
 *         """Doubles the size of the byteslice, and copies the old data towards
 *            the end of the new buffer (since we build the buffer backwards)."""
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.growByteBuffer", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":441
 *         self.buffer = buffer_2
 * 
 *     cdef Py_ssize_t Offset(self):             # <<<<<<<<<<<<<<
 *         """Offset relative to the end of the buffer."""
 *         return self.buffer_length - self.head
 */

static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_Offset(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Offset", 0);

  /* "flattr/cflattr/builder.pyx":443
 *     cdef Py_ssize_t Offset(self):
 *         """Offset relative to the end of the buffer."""
 *         return self.buffer_length - self.head             # <<<<<<<<<<<<<<
 * 
 *     cdef Pad(self, Py_ssize_t n):
 */
  __pyx_r = (__pyx_v_self->buffer_length - __pyx_v_self->head);
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":441
 *         self.buffer = buffer_2
 * 
 *     cdef Py_ssize_t Offset(self):             # <<<<<<<<<<<<<<
 *         """Offset relative to the end of the buffer."""
 *         return self.buffer_length - self.head
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":445
 *         return self.buffer_length - self.head
 * 
 *     cdef Pad(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         """Pad places zeros at the current offset."""
 *         cdef int i
 */

static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_Pad(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_n) {
  CYTHON_UNUSED int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("Pad", 0);

  /* "flattr/cflattr/builder.pyx":448
 *         """Pad places zeros at the current offset."""
 *         cdef int i
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             self.Place(0, &Fb_uint8_t)
 * 
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flattr/cflattr/builder.pyx":449
 *         cdef int i
 *         for i in range(n):
 *             self.Place(0, &Fb_uint8_t)             # <<<<<<<<<<<<<<
 * 
 *     cdef void Prep(self, Py_ssize_t size, Py_ssize_t additionalBytes):
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Place(__pyx_v_self, 0, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t));
  }

  /* "flattr/cflattr/builder.pyx":445
 *         return self.buffer_length - self.head
 * 
 *     cdef Pad(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         """Pad places zeros at the current offset."""
 *         cdef int i
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":451
 *             self.Place(0, &Fb_uint8_t)
 * 
 *     cdef void Prep(self, Py_ssize_t size, Py_ssize_t additionalBytes):             # <<<<<<<<<<<<<<
 *         """
 *         Prep prepares to write an element of `size` after `additional_bytes`
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Prep(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_size, Py_ssize_t __pyx_v_additionalBytes) {
  long __pyx_v_alignSize;
  long __pyx_v_updated_head;
  long __pyx_v_oldBufSize;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Prep", 0);

  /* "flattr/cflattr/builder.pyx":461
 *         cdef long alignSize, updated_head, oldBufSize
 *         # Track the biggest thing we've ever aligned to.
 *         if size > self.minalign:             # <<<<<<<<<<<<<<
 *             self.minalign = size
 * 
 */
  __pyx_t_1 = ((__pyx_v_size > __pyx_v_self->minalign) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":462
 *         # Track the biggest thing we've ever aligned to.
 *         if size > self.minalign:
 *             self.minalign = size             # <<<<<<<<<<<<<<
 * 
 *         # Find the amount of alignment needed such that `size` is properly
 */
    __pyx_v_self->minalign = __pyx_v_size;

    /* "flattr/cflattr/builder.pyx":461
 *         cdef long alignSize, updated_head, oldBufSize
 *         # Track the biggest thing we've ever aligned to.
 *         if size > self.minalign:             # <<<<<<<<<<<<<<
 *             self.minalign = size
 * 
 */
  }

  /* "flattr/cflattr/builder.pyx":466
 *         # Find the amount of alignment needed such that `size` is properly
 *         # aligned after `additionalBytes`:
 *         alignSize = (~(self.buffer_length - self.head + additionalBytes)) + 1             # <<<<<<<<<<<<<<
 *         alignSize &= (size - 1)
 * 
 */
  __pyx_v_alignSize = ((~((__pyx_v_self->buffer_length - __pyx_v_self->head) + __pyx_v_additionalBytes)) + 1);

  /* "flattr/cflattr/builder.pyx":467
 *         # aligned after `additionalBytes`:
 *         alignSize = (~(self.buffer_length - self.head + additionalBytes)) + 1
 *         alignSize &= (size - 1)             # <<<<<<<<<<<<<<
 * 
 *         # Reallocate the buffer if needed:
 */
  __pyx_v_alignSize = (__pyx_v_alignSize & (__pyx_v_size - 1));

  /* "flattr/cflattr/builder.pyx":470
 * 
 *         # Reallocate the buffer if needed:
 *         while self.head < alignSize+size+additionalBytes:             # <<<<<<<<<<<<<<
 *             oldBufSize = self.buffer_length
 *             self.growByteBuffer()
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_self->head < ((__pyx_v_alignSize + __pyx_v_size) + __pyx_v_additionalBytes)) != 0);
    if (!__pyx_t_1) break;

    /* "flattr/cflattr/builder.pyx":471
 *         # Reallocate the buffer if needed:
 *         while self.head < alignSize+size+additionalBytes:
 *             oldBufSize = self.buffer_length             # <<<<<<<<<<<<<<
 *             self.growByteBuffer()
 *             updated_head = self.head + self.buffer_length - oldBufSize
 */
    __pyx_t_2 = __pyx_v_self->buffer_length;
    __pyx_v_oldBufSize = __pyx_t_2;

    /* "flattr/cflattr/builder.pyx":472
 *         while self.head < alignSize+size+additionalBytes:
 *             oldBufSize = self.buffer_length
 *             self.growByteBuffer()             # <<<<<<<<<<<<<<
 *             updated_head = self.head + self.buffer_length - oldBufSize
 *             self.head = updated_head
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->growByteBuffer(__pyx_v_self);

    /* "flattr/cflattr/builder.pyx":473
 *             oldBufSize = self.buffer_length
 *             self.growByteBuffer()
 *             updated_head = self.head + self.buffer_length - oldBufSize             # <<<<<<<<<<<<<<
 *             self.head = updated_head
 *         self.Pad(alignSize)
 */
    __pyx_v_updated_head = ((__pyx_v_self->head + __pyx_v_self->buffer_length) - __pyx_v_oldBufSize);

    /* "flattr/cflattr/builder.pyx":474
 *             self.growByteBuffer()
 *             updated_head = self.head + self.buffer_length - oldBufSize
 *             self.head = updated_head             # <<<<<<<<<<<<<<
 *         self.Pad(alignSize)
 * 
 */
    __pyx_v_self->head = __pyx_v_updated_head;
  }

  /* "flattr/cflattr/builder.pyx":475
 *             updated_head = self.head + self.buffer_length - oldBufSize
 *             self.head = updated_head
 *         self.Pad(alignSize)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PrependSOffsetTRelative(self, int off):
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Pad(__pyx_v_self, __pyx_v_alignSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flattr/cflattr/builder.pyx":451
 *             self.Place(0, &Fb_uint8_t)
 * 
 *     cdef void Prep(self, Py_ssize_t size, Py_ssize_t additionalBytes):             # <<<<<<<<<<<<<<
 *         """
 *         Prep prepares to write an element of `size` after `additional_bytes`
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.Prep", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":477
 *         self.Pad(alignSize)
 * 
 *     cdef void PrependSOffsetTRelative(self, int off):             # <<<<<<<<<<<<<<
 *         """
 *         PrependSOffsetTRelative prepends an SOffsetT, relative to where it
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_off) {
  PyObject *__pyx_v_msg = NULL;
  Py_ssize_t __pyx_v_off2;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependSOffsetTRelative", 0);

  /* "flattr/cflattr/builder.pyx":484
 * 
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_int32_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t.bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":485
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_int32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 */
  __pyx_t_1 = ((!((__pyx_v_off <= ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self)) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":486
 *         self.Prep(Fb_int32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."             # <<<<<<<<<<<<<<
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_int32_t.bytewidth
 */
    __Pyx_INCREF(__pyx_kp_u_flatbuffers_Offset_arithmetic_er);
    __pyx_v_msg = __pyx_kp_u_flatbuffers_Offset_arithmetic_er;

    /* "flattr/cflattr/builder.pyx":487
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)             # <<<<<<<<<<<<<<
 *         off2 = self.Offset() - off + Fb_int32_t.bytewidth
 *         self.PlaceSOffsetT(off2)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OffsetArithmeticError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 487, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":485
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_int32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 */
  }

  /* "flattr/cflattr/builder.pyx":488
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_int32_t.bytewidth             # <<<<<<<<<<<<<<
 *         self.PlaceSOffsetT(off2)
 * 
 */
  __pyx_v_off2 = ((((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self) - __pyx_v_off) + __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":489
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_int32_t.bytewidth
 *         self.PlaceSOffsetT(off2)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void PrependUOffsetTRelative(self, Py_ssize_t off):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PlaceSOffsetT(__pyx_v_self, __pyx_v_off2);

  /* "flattr/cflattr/builder.pyx":477
 *         self.Pad(alignSize)
 * 
 *     cdef void PrependSOffsetTRelative(self, int off):             # <<<<<<<<<<<<<<
 *         """
 *         PrependSOffsetTRelative prepends an SOffsetT, relative to where it
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.PrependSOffsetTRelative", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":491
 *         self.PlaceSOffsetT(off2)
 * 
 *     cpdef void PrependUOffsetTRelative(self, Py_ssize_t off):             # <<<<<<<<<<<<<<
 *         """Prepends an unsigned offset into vector data, relative to where it
 *         will be written.
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_11PrependUOffsetTRelative(PyObject *__pyx_v_self, PyObject *__pyx_arg_off); /*proto*/
static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependUOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_off, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_msg = NULL;
  Py_ssize_t __pyx_v_off2;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUOffsetTRelative", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_PrependUOffsetTRelative); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6flattr_7cflattr_7builder_7Builder_11PrependUOffsetTRelative)) {
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_off); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flattr/cflattr/builder.pyx":497
 * 
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_uint32_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":498
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_uint32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 */
  __pyx_t_6 = ((!((__pyx_v_off <= ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self)) != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "flattr/cflattr/builder.pyx":499
 *         self.Prep(Fb_uint32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."             # <<<<<<<<<<<<<<
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_uint32_t.bytewidth
 */
    __Pyx_INCREF(__pyx_kp_u_flatbuffers_Offset_arithmetic_er);
    __pyx_v_msg = __pyx_kp_u_flatbuffers_Offset_arithmetic_er;

    /* "flattr/cflattr/builder.pyx":500
 *         if not (off <= self.Offset()):
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)             # <<<<<<<<<<<<<<
 *         off2 = self.Offset() - off + Fb_uint32_t.bytewidth
 *         self.PlaceUOffsetT(off2)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OffsetArithmeticError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 500, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":498
 *         # Ensure alignment is already done:
 *         self.Prep(Fb_uint32_t.bytewidth, 0)
 *         if not (off <= self.Offset()):             # <<<<<<<<<<<<<<
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 */
  }

  /* "flattr/cflattr/builder.pyx":501
 *             msg = "flatbuffers: Offset arithmetic error."
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_uint32_t.bytewidth             # <<<<<<<<<<<<<<
 *         self.PlaceUOffsetT(off2)
 * 
 */
  __pyx_v_off2 = ((((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self) - __pyx_v_off) + __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":502
 *             raise OffsetArithmeticError(msg)
 *         off2 = self.Offset() - off + Fb_uint32_t.bytewidth
 *         self.PlaceUOffsetT(off2)             # <<<<<<<<<<<<<<
 * 
 *     def StartVector(self, int elemSize, Py_ssize_t numElems, Py_ssize_t alignment):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PlaceUOffsetT(__pyx_v_self, __pyx_v_off2);

  /* "flattr/cflattr/builder.pyx":491
 *         self.PlaceSOffsetT(off2)
 * 
 *     cpdef void PrependUOffsetTRelative(self, Py_ssize_t off):             # <<<<<<<<<<<<<<
 *         """Prepends an unsigned offset into vector data, relative to where it
 *         will be written.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.PrependUOffsetTRelative", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_11PrependUOffsetTRelative(PyObject *__pyx_v_self, PyObject *__pyx_arg_off); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_10PrependUOffsetTRelative[] = "Prepends an unsigned offset into vector data, relative to where it\n        will be written.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_11PrependUOffsetTRelative(PyObject *__pyx_v_self, PyObject *__pyx_arg_off) {
  Py_ssize_t __pyx_v_off;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUOffsetTRelative (wrapper)", 0);
  assert(__pyx_arg_off); {
    __pyx_v_off = __Pyx_PyIndex_AsSsize_t(__pyx_arg_off); if (unlikely((__pyx_v_off == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUOffsetTRelative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_10PrependUOffsetTRelative(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((Py_ssize_t)__pyx_v_off));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_10PrependUOffsetTRelative(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_off) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUOffsetTRelative", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependUOffsetTRelative(__pyx_v_self, __pyx_v_off, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUOffsetTRelative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":504
 *         self.PlaceUOffsetT(off2)
 * 
 *     def StartVector(self, int elemSize, Py_ssize_t numElems, Py_ssize_t alignment):             # <<<<<<<<<<<<<<
 *         """
 *         StartVector initializes bookkeeping for writing a new vector.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_13StartVector(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_12StartVector[] = "\n        StartVector initializes bookkeeping for writing a new vector.\n\n        A vector has the following format:\n          - <UOffsetT: number of elements in this vector>\n          - <T: data>+, where T is the type of elements of this vector.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_13StartVector(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_elemSize;
  Py_ssize_t __pyx_v_numElems;
  Py_ssize_t __pyx_v_alignment;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("StartVector (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_elemSize,&__pyx_n_s_numElems,&__pyx_n_s_alignment,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_elemSize)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_numElems)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("StartVector", 1, 3, 3, 1); __PYX_ERR(0, 504, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alignment)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("StartVector", 1, 3, 3, 2); __PYX_ERR(0, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "StartVector") < 0)) __PYX_ERR(0, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_elemSize = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_elemSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
    __pyx_v_numElems = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_numElems == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
    __pyx_v_alignment = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_alignment == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("StartVector", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.StartVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_12StartVector(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_elemSize, __pyx_v_numElems, __pyx_v_alignment);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_12StartVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_elemSize, Py_ssize_t __pyx_v_numElems, Py_ssize_t __pyx_v_alignment) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("StartVector", 0);

  /* "flattr/cflattr/builder.pyx":513
 *         """
 * 
 *         self.assertNotNested()             # <<<<<<<<<<<<<<
 *         self.nested = True
 *         self.vectorNumElems = numElems
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNotNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 513, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":514
 * 
 *         self.assertNotNested()
 *         self.nested = True             # <<<<<<<<<<<<<<
 *         self.vectorNumElems = numElems
 *         self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)
 */
  __pyx_v_self->nested = 1;

  /* "flattr/cflattr/builder.pyx":515
 *         self.assertNotNested()
 *         self.nested = True
 *         self.vectorNumElems = numElems             # <<<<<<<<<<<<<<
 *         self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)
 *         self.Prep(alignment, elemSize*numElems)  # In case alignment > int.
 */
  __pyx_v_self->vectorNumElems = __pyx_v_numElems;

  /* "flattr/cflattr/builder.pyx":516
 *         self.nested = True
 *         self.vectorNumElems = numElems
 *         self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)             # <<<<<<<<<<<<<<
 *         self.Prep(alignment, elemSize*numElems)  # In case alignment > int.
 *         return self.Offset()
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth, (__pyx_v_elemSize * __pyx_v_numElems));

  /* "flattr/cflattr/builder.pyx":517
 *         self.vectorNumElems = numElems
 *         self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)
 *         self.Prep(alignment, elemSize*numElems)  # In case alignment > int.             # <<<<<<<<<<<<<<
 *         return self.Offset()
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_alignment, (__pyx_v_elemSize * __pyx_v_numElems));

  /* "flattr/cflattr/builder.pyx":518
 *         self.Prep(Fb_uint32_t.bytewidth, elemSize*numElems)
 *         self.Prep(alignment, elemSize*numElems)  # In case alignment > int.
 *         return self.Offset()             # <<<<<<<<<<<<<<
 * 
 *     cpdef Py_ssize_t EndVector(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":504
 *         self.PlaceUOffsetT(off2)
 * 
 *     def StartVector(self, int elemSize, Py_ssize_t numElems, Py_ssize_t alignment):             # <<<<<<<<<<<<<<
 *         """
 *         StartVector initializes bookkeeping for writing a new vector.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.StartVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":520
 *         return self.Offset()
 * 
 *     cpdef Py_ssize_t EndVector(self):             # <<<<<<<<<<<<<<
 *         """EndVector writes data necessary to finish vector construction."""
 * 
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_15EndVector(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_EndVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EndVector", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_EndVector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6flattr_7cflattr_7builder_7Builder_15EndVector)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flattr/cflattr/builder.pyx":523
 *         """EndVector writes data necessary to finish vector construction."""
 * 
 *         self.assertNested()             # <<<<<<<<<<<<<<
 *         ## @cond FLATBUFFERS_INTERNAL
 *         self.nested = False
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":525
 *         self.assertNested()
 *         ## @cond FLATBUFFERS_INTERNAL
 *         self.nested = False             # <<<<<<<<<<<<<<
 *         ## @endcond
 *         # we already made space for this, so write without PrependUint32
 */
  __pyx_v_self->nested = 0;

  /* "flattr/cflattr/builder.pyx":528
 *         ## @endcond
 *         # we already made space for this, so write without PrependUint32
 *         self.PlaceUOffsetT(self.vectorNumElems)             # <<<<<<<<<<<<<<
 *         return self.Offset()
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PlaceUOffsetT(__pyx_v_self, __pyx_v_self->vectorNumElems);

  /* "flattr/cflattr/builder.pyx":529
 *         # we already made space for this, so write without PrependUint32
 *         self.PlaceUOffsetT(self.vectorNumElems)
 *         return self.Offset()             # <<<<<<<<<<<<<<
 * 
 *     cpdef Py_ssize_t CreateString(self, str s, encoding='utf-8', errors='strict'):
 */
  __pyx_r = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":520
 *         return self.Offset()
 * 
 *     cpdef Py_ssize_t EndVector(self):             # <<<<<<<<<<<<<<
 *         """EndVector writes data necessary to finish vector construction."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.EndVector", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_15EndVector(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_14EndVector[] = "EndVector writes data necessary to finish vector construction.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_15EndVector(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EndVector (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_14EndVector(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_14EndVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EndVector", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_f_6flattr_7cflattr_7builder_7Builder_EndVector(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.EndVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":531
 *         return self.Offset()
 * 
 *     cpdef Py_ssize_t CreateString(self, str s, encoding='utf-8', errors='strict'):             # <<<<<<<<<<<<<<
 *         """CreateString writes a null-terminated byte string as a vector."""
 *         cdef bytes x
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_17CreateString(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_CreateString(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch, struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString *__pyx_optional_args) {
  PyObject *__pyx_v_encoding = ((PyObject *)__pyx_kp_u_utf_8);
  PyObject *__pyx_v_errors = ((PyObject *)__pyx_n_u_strict);
  PyObject *__pyx_v_x = 0;
  Py_ssize_t __pyx_v_l;
  char *__pyx_v_c_string;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  char *__pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("CreateString", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_encoding = __pyx_optional_args->encoding;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_errors = __pyx_optional_args->errors;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_CreateString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6flattr_7cflattr_7builder_7Builder_17CreateString)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_s, __pyx_v_encoding, __pyx_v_errors};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_s, __pyx_v_encoding, __pyx_v_errors};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 531, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_s);
          __Pyx_GIVEREF(__pyx_v_s);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_s);
          __Pyx_INCREF(__pyx_v_encoding);
          __Pyx_GIVEREF(__pyx_v_encoding);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_encoding);
          __Pyx_INCREF(__pyx_v_errors);
          __Pyx_GIVEREF(__pyx_v_errors);
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_errors);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flattr/cflattr/builder.pyx":536
 *         cdef Py_ssize_t l
 * 
 *         self.assertNotNested()             # <<<<<<<<<<<<<<
 *         ## @cond FLATBUFFERS_INTERNAL
 *         self.nested = True
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNotNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":538
 *         self.assertNotNested()
 *         ## @cond FLATBUFFERS_INTERNAL
 *         self.nested = True             # <<<<<<<<<<<<<<
 *         ## @endcond
 * 
 */
  __pyx_v_self->nested = 1;

  /* "flattr/cflattr/builder.pyx":541
 *         ## @endcond
 * 
 *         x = s.encode(encoding)             # <<<<<<<<<<<<<<
 *         l = len(x)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_encoding);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 541, __pyx_L1_error)
  __pyx_v_x = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flattr/cflattr/builder.pyx":542
 * 
 *         x = s.encode(encoding)
 *         l = len(x)             # <<<<<<<<<<<<<<
 * 
 *         self.Prep(Fb_uint32_t.bytewidth, (l+1)*Fb_uint8_t.bytewidth)
 */
  if (unlikely(__pyx_v_x == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 542, __pyx_L1_error)
  }
  __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_x); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 542, __pyx_L1_error)
  __pyx_v_l = __pyx_t_7;

  /* "flattr/cflattr/builder.pyx":544
 *         l = len(x)
 * 
 *         self.Prep(Fb_uint32_t.bytewidth, (l+1)*Fb_uint8_t.bytewidth)             # <<<<<<<<<<<<<<
 *         self.Place(0, &Fb_uint8_t)
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth, ((__pyx_v_l + 1) * __pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t.bytewidth));

  /* "flattr/cflattr/builder.pyx":545
 * 
 *         self.Prep(Fb_uint32_t.bytewidth, (l+1)*Fb_uint8_t.bytewidth)
 *         self.Place(0, &Fb_uint8_t)             # <<<<<<<<<<<<<<
 * 
 *         self.head = self.head - l
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Place(__pyx_v_self, 0, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t));

  /* "flattr/cflattr/builder.pyx":547
 *         self.Place(0, &Fb_uint8_t)
 * 
 *         self.head = self.head - l             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_string = x
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_l);

  /* "flattr/cflattr/builder.pyx":549
 *         self.head = self.head - l
 * 
 *         cdef char* c_string = x             # <<<<<<<<<<<<<<
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)
 */
  if (unlikely(__pyx_v_x == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_x); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_v_c_string = __pyx_t_8;

  /* "flattr/cflattr/builder.pyx":551
 *         cdef char* c_string = x
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)             # <<<<<<<<<<<<<<
 *         self.vectorNumElems = l
 * 
 */
  (void)(memcpy(((void *)(&(__pyx_v_self->buffer[__pyx_v_self->head]))), __pyx_v_c_string, __pyx_v_l));

  /* "flattr/cflattr/builder.pyx":552
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)
 *         self.vectorNumElems = l             # <<<<<<<<<<<<<<
 * 
 *         return self.EndVector()
 */
  __pyx_v_self->vectorNumElems = __pyx_v_l;

  /* "flattr/cflattr/builder.pyx":554
 *         self.vectorNumElems = l
 * 
 *         return self.EndVector()             # <<<<<<<<<<<<<<
 * 
 *     cpdef Py_ssize_t CreateByteVector(self, bytes x):
 */
  __pyx_r = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->EndVector(__pyx_v_self, 0);
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":531
 *         return self.Offset()
 * 
 *     cpdef Py_ssize_t CreateString(self, str s, encoding='utf-8', errors='strict'):             # <<<<<<<<<<<<<<
 *         """CreateString writes a null-terminated byte string as a vector."""
 *         cdef bytes x
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.CreateString", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_17CreateString(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_16CreateString[] = "CreateString writes a null-terminated byte string as a vector.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_17CreateString(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_errors = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("CreateString (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_s,&__pyx_n_s_encoding,&__pyx_n_s_errors,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_kp_u_utf_8);
    values[2] = ((PyObject *)__pyx_n_u_strict);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_encoding);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_errors);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "CreateString") < 0)) __PYX_ERR(0, 531, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_s = ((PyObject*)values[0]);
    __pyx_v_encoding = values[1];
    __pyx_v_errors = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("CreateString", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 531, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.CreateString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyUnicode_Type), 1, "s", 1))) __PYX_ERR(0, 531, __pyx_L1_error)
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_16CreateString(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_s, __pyx_v_encoding, __pyx_v_errors);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_16CreateString(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_encoding, PyObject *__pyx_v_errors) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("CreateString", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.encoding = __pyx_v_encoding;
  __pyx_t_2.errors = __pyx_v_errors;
  __pyx_t_1 = __pyx_vtabptr_6flattr_7cflattr_7builder_Builder->CreateString(__pyx_v_self, __pyx_v_s, 1, &__pyx_t_2); 
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.CreateString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":556
 *         return self.EndVector()
 * 
 *     cpdef Py_ssize_t CreateByteVector(self, bytes x):             # <<<<<<<<<<<<<<
 *         """CreateString writes a byte vector."""
 *         self.assertNotNested()
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_19CreateByteVector(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder_CreateByteVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_l = 0;
  char *__pyx_v_c_string;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  char *__pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("CreateByteVector", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_CreateByteVector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6flattr_7cflattr_7builder_7Builder_19CreateByteVector)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_x) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 556, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flattr/cflattr/builder.pyx":558
 *     cpdef Py_ssize_t CreateByteVector(self, bytes x):
 *         """CreateString writes a byte vector."""
 *         self.assertNotNested()             # <<<<<<<<<<<<<<
 *         self.nested = True
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNotNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":559
 *         """CreateString writes a byte vector."""
 *         self.assertNotNested()
 *         self.nested = True             # <<<<<<<<<<<<<<
 * 
 *         cdef l = len(x)
 */
  __pyx_v_self->nested = 1;

  /* "flattr/cflattr/builder.pyx":561
 *         self.nested = True
 * 
 *         cdef l = len(x)             # <<<<<<<<<<<<<<
 * 
 *         self.Prep(Fb_uint32_t.bytewidth, l*Fb_uint8_t.bytewidth)
 */
  if (unlikely(__pyx_v_x == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 561, __pyx_L1_error)
  }
  __pyx_t_5 = PyBytes_GET_SIZE(__pyx_v_x); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 561, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_l = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flattr/cflattr/builder.pyx":563
 *         cdef l = len(x)
 * 
 *         self.Prep(Fb_uint32_t.bytewidth, l*Fb_uint8_t.bytewidth)             # <<<<<<<<<<<<<<
 * 
 *         self.head = self.head - l
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t.bytewidth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_l, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth, __pyx_t_5);

  /* "flattr/cflattr/builder.pyx":565
 *         self.Prep(Fb_uint32_t.bytewidth, l*Fb_uint8_t.bytewidth)
 * 
 *         self.head = self.head - l             # <<<<<<<<<<<<<<
 * 
 *         cdef char* c_string = x
 */
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->head); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_2, __pyx_v_l); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->head = __pyx_t_5;

  /* "flattr/cflattr/builder.pyx":567
 *         self.head = self.head - l
 * 
 *         cdef char* c_string = x             # <<<<<<<<<<<<<<
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)
 */
  if (unlikely(__pyx_v_x == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 567, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_x); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
  __pyx_v_c_string = __pyx_t_6;

  /* "flattr/cflattr/builder.pyx":569
 *         cdef char* c_string = x
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)             # <<<<<<<<<<<<<<
 *         self.vectorNumElems = l
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_l); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 569, __pyx_L1_error)
  (void)(memcpy(((void *)(&(__pyx_v_self->buffer[__pyx_v_self->head]))), __pyx_v_c_string, __pyx_t_7));

  /* "flattr/cflattr/builder.pyx":570
 * 
 *         memcpy(<void*>&self.buffer[self.head], c_string, l)
 *         self.vectorNumElems = l             # <<<<<<<<<<<<<<
 * 
 *         return self.EndVector()
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_l); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L1_error)
  __pyx_v_self->vectorNumElems = __pyx_t_5;

  /* "flattr/cflattr/builder.pyx":572
 *         self.vectorNumElems = l
 * 
 *         return self.EndVector()             # <<<<<<<<<<<<<<
 * 
 *     cdef void assertNested(self) except *:
 */
  __pyx_r = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->EndVector(__pyx_v_self, 0);
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":556
 *         return self.EndVector()
 * 
 *     cpdef Py_ssize_t CreateByteVector(self, bytes x):             # <<<<<<<<<<<<<<
 *         """CreateString writes a byte vector."""
 *         self.assertNotNested()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.CreateByteVector", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_19CreateByteVector(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_18CreateByteVector[] = "CreateString writes a byte vector.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_19CreateByteVector(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("CreateByteVector (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), (&PyBytes_Type), 1, "x", 1))) __PYX_ERR(0, 556, __pyx_L1_error)
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_18CreateByteVector(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject*)__pyx_v_x));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_18CreateByteVector(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("CreateByteVector", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_f_6flattr_7cflattr_7builder_7Builder_CreateByteVector(__pyx_v_self, __pyx_v_x, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.CreateByteVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":574
 *         return self.EndVector()
 * 
 *     cdef void assertNested(self) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Check that we are in the process of building an object.
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertNested(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assertNested", 0);

  /* "flattr/cflattr/builder.pyx":579
 *         """
 * 
 *         if not self.nested:             # <<<<<<<<<<<<<<
 *             raise IsNotNestedError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->nested != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":580
 * 
 *         if not self.nested:
 *             raise IsNotNestedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef void assertNotNested(self) except *:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_IsNotNestedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 580, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":579
 *         """
 * 
 *         if not self.nested:             # <<<<<<<<<<<<<<
 *             raise IsNotNestedError()
 * 
 */
  }

  /* "flattr/cflattr/builder.pyx":574
 *         return self.EndVector()
 * 
 *     cdef void assertNested(self) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Check that we are in the process of building an object.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.assertNested", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":582
 *             raise IsNotNestedError()
 * 
 *     cdef void assertNotNested(self) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Check that no other objects are being built while making this
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertNotNested(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assertNotNested", 0);

  /* "flattr/cflattr/builder.pyx":588
 *         """
 * 
 *         if self.nested:             # <<<<<<<<<<<<<<
 *             raise IsNestedError()
 * 
 */
  __pyx_t_1 = (__pyx_v_self->nested != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":589
 * 
 *         if self.nested:
 *             raise IsNestedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef void assertStructIsInline(self, Py_ssize_t obj) except *:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_IsNestedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 589, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":588
 *         """
 * 
 *         if self.nested:             # <<<<<<<<<<<<<<
 *             raise IsNestedError()
 * 
 */
  }

  /* "flattr/cflattr/builder.pyx":582
 *             raise IsNotNestedError()
 * 
 *     cdef void assertNotNested(self) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Check that no other objects are being built while making this
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.assertNotNested", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":591
 *             raise IsNestedError()
 * 
 *     cdef void assertStructIsInline(self, Py_ssize_t obj) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Structs are always stored inline, so need to be created right
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_assertStructIsInline(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_obj) {
  PyObject *__pyx_v_msg = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assertStructIsInline", 0);

  /* "flattr/cflattr/builder.pyx":598
 *         """
 * 
 *         if obj != self.Offset():             # <<<<<<<<<<<<<<
 *             msg = ("flatbuffers: Tried to write a Struct at an Offset that "
 *                    "is different from the current Offset of the Builder.")
 */
  __pyx_t_1 = ((__pyx_v_obj != ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":599
 * 
 *         if obj != self.Offset():
 *             msg = ("flatbuffers: Tried to write a Struct at an Offset that "             # <<<<<<<<<<<<<<
 *                    "is different from the current Offset of the Builder.")
 *             raise StructIsNotInlineError(msg)
 */
    __Pyx_INCREF(__pyx_kp_u_flatbuffers_Tried_to_write_a_Str);
    __pyx_v_msg = __pyx_kp_u_flatbuffers_Tried_to_write_a_Str;

    /* "flattr/cflattr/builder.pyx":601
 *             msg = ("flatbuffers: Tried to write a Struct at an Offset that "
 *                    "is different from the current Offset of the Builder.")
 *             raise StructIsNotInlineError(msg)             # <<<<<<<<<<<<<<
 * 
 *     cdef void Slot(self, Py_ssize_t slotnum) except *:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StructIsNotInlineError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_msg);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 601, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":598
 *         """
 * 
 *         if obj != self.Offset():             # <<<<<<<<<<<<<<
 *             msg = ("flatbuffers: Tried to write a Struct at an Offset that "
 *                    "is different from the current Offset of the Builder.")
 */
  }

  /* "flattr/cflattr/builder.pyx":591
 *             raise IsNestedError()
 * 
 *     cdef void assertStructIsInline(self, Py_ssize_t obj) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Structs are always stored inline, so need to be created right
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.assertStructIsInline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":603
 *             raise StructIsNotInlineError(msg)
 * 
 *     cdef void Slot(self, Py_ssize_t slotnum) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Slot sets the vtable key `voffset` to the current location in the
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_slotnum) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Slot", 0);

  /* "flattr/cflattr/builder.pyx":609
 * 
 *         """
 *         self.assertNested()             # <<<<<<<<<<<<<<
 *         if slotnum >= self.current_vtable_length:
 *             raise IndexError()
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertNested(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)

  /* "flattr/cflattr/builder.pyx":610
 *         """
 *         self.assertNested()
 *         if slotnum >= self.current_vtable_length:             # <<<<<<<<<<<<<<
 *             raise IndexError()
 *         self.current_vtable[slotnum] = self.Offset()
 */
  __pyx_t_1 = ((__pyx_v_slotnum >= __pyx_v_self->current_vtable_length) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flattr/cflattr/builder.pyx":611
 *         self.assertNested()
 *         if slotnum >= self.current_vtable_length:
 *             raise IndexError()             # <<<<<<<<<<<<<<
 *         self.current_vtable[slotnum] = self.Offset()
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_IndexError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 611, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":610
 *         """
 *         self.assertNested()
 *         if slotnum >= self.current_vtable_length:             # <<<<<<<<<<<<<<
 *             raise IndexError()
 *         self.current_vtable[slotnum] = self.Offset()
 */
  }

  /* "flattr/cflattr/builder.pyx":612
 *         if slotnum >= self.current_vtable_length:
 *             raise IndexError()
 *         self.current_vtable[slotnum] = self.Offset()             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t __Finish(self, Py_ssize_t rootTable, bint sizePrefix):
 */
  (__pyx_v_self->current_vtable[__pyx_v_slotnum]) = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Offset(__pyx_v_self);

  /* "flattr/cflattr/builder.pyx":603
 *             raise StructIsNotInlineError(msg)
 * 
 *     cdef void Slot(self, Py_ssize_t slotnum) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Slot sets the vtable key `voffset` to the current location in the
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":614
 *         self.current_vtable[slotnum] = self.Offset()
 * 
 *     cdef Py_ssize_t __Finish(self, Py_ssize_t rootTable, bint sizePrefix):             # <<<<<<<<<<<<<<
 *         """Finish finalizes a buffer, pointing to the given `rootTable`."""
 *         cdef Py_ssize_t prepSize, size
 */

static Py_ssize_t __pyx_f_6flattr_7cflattr_7builder_7Builder___Finish(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_rootTable, int __pyx_v_sizePrefix) {
  Py_ssize_t __pyx_v_prepSize;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Finish", 0);

  /* "flattr/cflattr/builder.pyx":618
 *         cdef Py_ssize_t prepSize, size
 * 
 *         prepSize = Fb_uint32_t.bytewidth             # <<<<<<<<<<<<<<
 *         if sizePrefix:
 *             prepSize += Fb_int32_t.bytewidth
 */
  __pyx_t_1 = __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth;
  __pyx_v_prepSize = __pyx_t_1;

  /* "flattr/cflattr/builder.pyx":619
 * 
 *         prepSize = Fb_uint32_t.bytewidth
 *         if sizePrefix:             # <<<<<<<<<<<<<<
 *             prepSize += Fb_int32_t.bytewidth
 * 
 */
  __pyx_t_2 = (__pyx_v_sizePrefix != 0);
  if (__pyx_t_2) {

    /* "flattr/cflattr/builder.pyx":620
 *         prepSize = Fb_uint32_t.bytewidth
 *         if sizePrefix:
 *             prepSize += Fb_int32_t.bytewidth             # <<<<<<<<<<<<<<
 * 
 *         self.Prep(self.minalign, prepSize)
 */
    __pyx_v_prepSize = (__pyx_v_prepSize + __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t.bytewidth);

    /* "flattr/cflattr/builder.pyx":619
 * 
 *         prepSize = Fb_uint32_t.bytewidth
 *         if sizePrefix:             # <<<<<<<<<<<<<<
 *             prepSize += Fb_int32_t.bytewidth
 * 
 */
  }

  /* "flattr/cflattr/builder.pyx":622
 *             prepSize += Fb_int32_t.bytewidth
 * 
 *         self.Prep(self.minalign, prepSize)             # <<<<<<<<<<<<<<
 *         self.PrependUOffsetTRelative(rootTable)
 *         if sizePrefix:
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_self->minalign, __pyx_v_prepSize);

  /* "flattr/cflattr/builder.pyx":623
 * 
 *         self.Prep(self.minalign, prepSize)
 *         self.PrependUOffsetTRelative(rootTable)             # <<<<<<<<<<<<<<
 *         if sizePrefix:
 *             size = self.buffer_length - self.head
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependUOffsetTRelative(__pyx_v_self, __pyx_v_rootTable, 0);

  /* "flattr/cflattr/builder.pyx":624
 *         self.Prep(self.minalign, prepSize)
 *         self.PrependUOffsetTRelative(rootTable)
 *         if sizePrefix:             # <<<<<<<<<<<<<<
 *             size = self.buffer_length - self.head
 *             self.PrependInt32(size)
 */
  __pyx_t_2 = (__pyx_v_sizePrefix != 0);
  if (__pyx_t_2) {

    /* "flattr/cflattr/builder.pyx":625
 *         self.PrependUOffsetTRelative(rootTable)
 *         if sizePrefix:
 *             size = self.buffer_length - self.head             # <<<<<<<<<<<<<<
 *             self.PrependInt32(size)
 *         self.finished = True
 */
    __pyx_v_size = (__pyx_v_self->buffer_length - __pyx_v_self->head);

    /* "flattr/cflattr/builder.pyx":626
 *         if sizePrefix:
 *             size = self.buffer_length - self.head
 *             self.PrependInt32(size)             # <<<<<<<<<<<<<<
 *         self.finished = True
 *         return self.head
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependInt32(__pyx_v_self, __pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flattr/cflattr/builder.pyx":624
 *         self.Prep(self.minalign, prepSize)
 *         self.PrependUOffsetTRelative(rootTable)
 *         if sizePrefix:             # <<<<<<<<<<<<<<
 *             size = self.buffer_length - self.head
 *             self.PrependInt32(size)
 */
  }

  /* "flattr/cflattr/builder.pyx":627
 *             size = self.buffer_length - self.head
 *             self.PrependInt32(size)
 *         self.finished = True             # <<<<<<<<<<<<<<
 *         return self.head
 * 
 */
  __pyx_v_self->finished = 1;

  /* "flattr/cflattr/builder.pyx":628
 *             self.PrependInt32(size)
 *         self.finished = True
 *         return self.head             # <<<<<<<<<<<<<<
 * 
 *     def Finish(self, int rootTable):
 */
  __pyx_r = __pyx_v_self->head;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":614
 *         self.current_vtable[slotnum] = self.Offset()
 * 
 *     cdef Py_ssize_t __Finish(self, Py_ssize_t rootTable, bint sizePrefix):             # <<<<<<<<<<<<<<
 *         """Finish finalizes a buffer, pointing to the given `rootTable`."""
 *         cdef Py_ssize_t prepSize, size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.__Finish", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":630
 *         return self.head
 * 
 *     def Finish(self, int rootTable):             # <<<<<<<<<<<<<<
 *         """Finish finalizes a buffer, pointing to the given `rootTable`."""
 *         return self.__Finish(rootTable, False)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_21Finish(PyObject *__pyx_v_self, PyObject *__pyx_arg_rootTable); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_20Finish[] = "Finish finalizes a buffer, pointing to the given `rootTable`.";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_21Finish(PyObject *__pyx_v_self, PyObject *__pyx_arg_rootTable) {
  int __pyx_v_rootTable;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Finish (wrapper)", 0);
  assert(__pyx_arg_rootTable); {
    __pyx_v_rootTable = __Pyx_PyInt_As_int(__pyx_arg_rootTable); if (unlikely((__pyx_v_rootTable == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.Finish", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_20Finish(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((int)__pyx_v_rootTable));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_20Finish(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int __pyx_v_rootTable) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Finish", 0);

  /* "flattr/cflattr/builder.pyx":632
 *     def Finish(self, int rootTable):
 *         """Finish finalizes a buffer, pointing to the given `rootTable`."""
 *         return self.__Finish(rootTable, False)             # <<<<<<<<<<<<<<
 * 
 *     def FinishSizePrefixed(self, Py_ssize_t rootTable):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->__pyx___Finish(__pyx_v_self, __pyx_v_rootTable, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":630
 *         return self.head
 * 
 *     def Finish(self, int rootTable):             # <<<<<<<<<<<<<<
 *         """Finish finalizes a buffer, pointing to the given `rootTable`."""
 *         return self.__Finish(rootTable, False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.Finish", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":634
 *         return self.__Finish(rootTable, False)
 * 
 *     def FinishSizePrefixed(self, Py_ssize_t rootTable):             # <<<<<<<<<<<<<<
 *         """
 *         Finish finalizes a buffer, pointing to the given `rootTable`,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_23FinishSizePrefixed(PyObject *__pyx_v_self, PyObject *__pyx_arg_rootTable); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_22FinishSizePrefixed[] = "\n        Finish finalizes a buffer, pointing to the given `rootTable`,\n        with the size prefixed.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_23FinishSizePrefixed(PyObject *__pyx_v_self, PyObject *__pyx_arg_rootTable) {
  Py_ssize_t __pyx_v_rootTable;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FinishSizePrefixed (wrapper)", 0);
  assert(__pyx_arg_rootTable); {
    __pyx_v_rootTable = __Pyx_PyIndex_AsSsize_t(__pyx_arg_rootTable); if (unlikely((__pyx_v_rootTable == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 634, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.FinishSizePrefixed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_22FinishSizePrefixed(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((Py_ssize_t)__pyx_v_rootTable));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_22FinishSizePrefixed(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_rootTable) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("FinishSizePrefixed", 0);

  /* "flattr/cflattr/builder.pyx":639
 *         with the size prefixed.
 *         """
 *         return self.__Finish(rootTable, True)             # <<<<<<<<<<<<<<
 * 
 *     cdef void Prepend(self, FlatbufferType* type, int64_t off):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->__pyx___Finish(__pyx_v_self, __pyx_v_rootTable, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flattr/cflattr/builder.pyx":634
 *         return self.__Finish(rootTable, False)
 * 
 *     def FinishSizePrefixed(self, Py_ssize_t rootTable):             # <<<<<<<<<<<<<<
 *         """
 *         Finish finalizes a buffer, pointing to the given `rootTable`,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.FinishSizePrefixed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":641
 *         return self.__Finish(rootTable, True)
 * 
 *     cdef void Prepend(self, FlatbufferType* type, int64_t off):             # <<<<<<<<<<<<<<
 *         self.Prep(type.bytewidth, 0)
 *         self.Place(off, type)
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Prepend(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_type, int64_t __pyx_v_off) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Prepend", 0);

  /* "flattr/cflattr/builder.pyx":642
 * 
 *     cdef void Prepend(self, FlatbufferType* type, int64_t off):
 *         self.Prep(type.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         self.Place(off, type)
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_type->bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":643
 *     cdef void Prepend(self, FlatbufferType* type, int64_t off):
 *         self.Prep(type.bytewidth, 0)
 *         self.Place(off, type)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Place(__pyx_v_self, __pyx_v_off, __pyx_v_type);

  /* "flattr/cflattr/builder.pyx":641
 *         return self.__Finish(rootTable, True)
 * 
 *     cdef void Prepend(self, FlatbufferType* type, int64_t off):             # <<<<<<<<<<<<<<
 *         self.Prep(type.bytewidth, 0)
 *         self.Place(off, type)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":645
 *         self.Place(off, type)
 * 
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prepend(type, x)
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_type, Py_ssize_t __pyx_v_o, int64_t __pyx_v_x, int64_t __pyx_v_d) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependSlot", 0);

  /* "flattr/cflattr/builder.pyx":646
 * 
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prepend(type, x)
 *             self.Slot(o)
 */
  __pyx_t_1 = ((__pyx_v_x != __pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":647
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):
 *         if x != d:
 *             self.Prepend(type, x)             # <<<<<<<<<<<<<<
 *             self.Slot(o)
 * 
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, __pyx_v_type, __pyx_v_x);

    /* "flattr/cflattr/builder.pyx":648
 *         if x != d:
 *             self.Prepend(type, x)
 *             self.Slot(o)             # <<<<<<<<<<<<<<
 * 
 *     def PrependBoolSlot(self, Py_ssize_t o, bint x, bint d):
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Slot(__pyx_v_self, __pyx_v_o); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":646
 * 
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prepend(type, x)
 *             self.Slot(o)
 */
  }

  /* "flattr/cflattr/builder.pyx":645
 *         self.Place(off, type)
 * 
 *     cdef void PrependSlot(self, FlatbufferType* type, Py_ssize_t o, int64_t x, int64_t d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prepend(type, x)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("flattr.cflattr.builder.Builder.PrependSlot", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":650
 *             self.Slot(o)
 * 
 *     def PrependBoolSlot(self, Py_ssize_t o, bint x, bint d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_bool_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_25PrependBoolSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_25PrependBoolSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int __pyx_v_x;
  int __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependBoolSlot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependBoolSlot", 1, 3, 3, 1); __PYX_ERR(0, 650, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependBoolSlot", 1, 3, 3, 2); __PYX_ERR(0, 650, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependBoolSlot") < 0)) __PYX_ERR(0, 650, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_d == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependBoolSlot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 650, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependBoolSlot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_24PrependBoolSlot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_24PrependBoolSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int __pyx_v_x, int __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependBoolSlot", 0);

  /* "flattr/cflattr/builder.pyx":651
 * 
 *     def PrependBoolSlot(self, Py_ssize_t o, bint x, bint d):
 *         self.PrependSlot(&Fb_bool_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependByteSlot(self, Py_ssize_t o, int x, int d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_bool_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":650
 *             self.Slot(o)
 * 
 *     def PrependBoolSlot(self, Py_ssize_t o, bint x, bint d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_bool_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":653
 *         self.PrependSlot(&Fb_bool_t, o, x, d)
 * 
 *     def PrependByteSlot(self, Py_ssize_t o, int x, int d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_27PrependByteSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_27PrependByteSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int __pyx_v_x;
  int __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependByteSlot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependByteSlot", 1, 3, 3, 1); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependByteSlot", 1, 3, 3, 2); __PYX_ERR(0, 653, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependByteSlot") < 0)) __PYX_ERR(0, 653, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_d == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependByteSlot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 653, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependByteSlot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_26PrependByteSlot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_26PrependByteSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int __pyx_v_x, int __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependByteSlot", 0);

  /* "flattr/cflattr/builder.pyx":654
 * 
 *     def PrependByteSlot(self, Py_ssize_t o, int x, int d):
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint8Slot(self, Py_ssize_t o, uint8_t x, uint8_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":653
 *         self.PrependSlot(&Fb_bool_t, o, x, d)
 * 
 *     def PrependByteSlot(self, Py_ssize_t o, int x, int d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":656
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 *     def PrependUint8Slot(self, Py_ssize_t o, uint8_t x, uint8_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_29PrependUint8Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_29PrependUint8Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  uint8_t __pyx_v_x;
  uint8_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint8Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint8Slot", 1, 3, 3, 1); __PYX_ERR(0, 656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint8Slot", 1, 3, 3, 2); __PYX_ERR(0, 656, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependUint8Slot") < 0)) __PYX_ERR(0, 656, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint8_t(values[1]); if (unlikely((__pyx_v_x == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_uint8_t(values[2]); if (unlikely((__pyx_v_d == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependUint8Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 656, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint8Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_28PrependUint8Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_28PrependUint8Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint8_t __pyx_v_x, uint8_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint8Slot", 0);

  /* "flattr/cflattr/builder.pyx":657
 * 
 *     def PrependUint8Slot(self, Py_ssize_t o, uint8_t x, uint8_t d):
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint16Slot(self, Py_ssize_t o, uint16_t x, uint16_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":656
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 *     def PrependUint8Slot(self, Py_ssize_t o, uint8_t x, uint8_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":659
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 *     def PrependUint16Slot(self, Py_ssize_t o, uint16_t x, uint16_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint16_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_31PrependUint16Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_31PrependUint16Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  uint16_t __pyx_v_x;
  uint16_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint16Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint16Slot", 1, 3, 3, 1); __PYX_ERR(0, 659, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint16Slot", 1, 3, 3, 2); __PYX_ERR(0, 659, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependUint16Slot") < 0)) __PYX_ERR(0, 659, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint16_t(values[1]); if (unlikely((__pyx_v_x == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_uint16_t(values[2]); if (unlikely((__pyx_v_d == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependUint16Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 659, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint16Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_30PrependUint16Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_30PrependUint16Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint16_t __pyx_v_x, uint16_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint16Slot", 0);

  /* "flattr/cflattr/builder.pyx":660
 * 
 *     def PrependUint16Slot(self, Py_ssize_t o, uint16_t x, uint16_t d):
 *         self.PrependSlot(&Fb_uint16_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint32Slot(self, Py_ssize_t o, uint32_t x, uint32_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":659
 *         self.PrependSlot(&Fb_uint8_t, o, x, d)
 * 
 *     def PrependUint16Slot(self, Py_ssize_t o, uint16_t x, uint16_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint16_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":662
 *         self.PrependSlot(&Fb_uint16_t, o, x, d)
 * 
 *     def PrependUint32Slot(self, Py_ssize_t o, uint32_t x, uint32_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint32_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_33PrependUint32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_33PrependUint32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  uint32_t __pyx_v_x;
  uint32_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint32Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint32Slot", 1, 3, 3, 1); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint32Slot", 1, 3, 3, 2); __PYX_ERR(0, 662, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependUint32Slot") < 0)) __PYX_ERR(0, 662, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_uint32_t(values[2]); if (unlikely((__pyx_v_d == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependUint32Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 662, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint32Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_32PrependUint32Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_32PrependUint32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint32_t __pyx_v_x, uint32_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint32Slot", 0);

  /* "flattr/cflattr/builder.pyx":663
 * 
 *     def PrependUint32Slot(self, Py_ssize_t o, uint32_t x, uint32_t d):
 *         self.PrependSlot(&Fb_uint32_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint64Slot(self, Py_ssize_t o, uint64_t x, uint64_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":662
 *         self.PrependSlot(&Fb_uint16_t, o, x, d)
 * 
 *     def PrependUint32Slot(self, Py_ssize_t o, uint32_t x, uint32_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint32_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":665
 *         self.PrependSlot(&Fb_uint32_t, o, x, d)
 * 
 *     def PrependUint64Slot(self, Py_ssize_t o, uint64_t x, uint64_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint64_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_35PrependUint64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_35PrependUint64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  uint64_t __pyx_v_x;
  uint64_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint64Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint64Slot", 1, 3, 3, 1); __PYX_ERR(0, 665, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUint64Slot", 1, 3, 3, 2); __PYX_ERR(0, 665, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependUint64Slot") < 0)) __PYX_ERR(0, 665, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(values[1]); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_uint64_t(values[2]); if (unlikely((__pyx_v_d == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependUint64Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 665, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint64Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_34PrependUint64Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_34PrependUint64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, uint64_t __pyx_v_x, uint64_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint64Slot", 0);

  /* "flattr/cflattr/builder.pyx":666
 * 
 *     def PrependUint64Slot(self, Py_ssize_t o, uint64_t x, uint64_t d):
 *         self.PrependSlot(&Fb_uint64_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt8Slot(self, Py_ssize_t o, int8_t x, int8_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":665
 *         self.PrependSlot(&Fb_uint32_t, o, x, d)
 * 
 *     def PrependUint64Slot(self, Py_ssize_t o, uint64_t x, uint64_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_uint64_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":668
 *         self.PrependSlot(&Fb_uint64_t, o, x, d)
 * 
 *     def PrependInt8Slot(self, Py_ssize_t o, int8_t x, int8_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int8_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_37PrependInt8Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_37PrependInt8Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int8_t __pyx_v_x;
  int8_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt8Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt8Slot", 1, 3, 3, 1); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt8Slot", 1, 3, 3, 2); __PYX_ERR(0, 668, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependInt8Slot") < 0)) __PYX_ERR(0, 668, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int8_t(values[1]); if (unlikely((__pyx_v_x == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_int8_t(values[2]); if (unlikely((__pyx_v_d == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependInt8Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 668, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt8Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_36PrependInt8Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_36PrependInt8Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int8_t __pyx_v_x, int8_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt8Slot", 0);

  /* "flattr/cflattr/builder.pyx":669
 * 
 *     def PrependInt8Slot(self, Py_ssize_t o, int8_t x, int8_t d):
 *         self.PrependSlot(&Fb_int8_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt16Slot(self, Py_ssize_t o, int16_t x, int16_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int8_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":668
 *         self.PrependSlot(&Fb_uint64_t, o, x, d)
 * 
 *     def PrependInt8Slot(self, Py_ssize_t o, int8_t x, int8_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int8_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":671
 *         self.PrependSlot(&Fb_int8_t, o, x, d)
 * 
 *     def PrependInt16Slot(self, Py_ssize_t o, int16_t x, int16_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int16_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_39PrependInt16Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_39PrependInt16Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int16_t __pyx_v_x;
  int16_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt16Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt16Slot", 1, 3, 3, 1); __PYX_ERR(0, 671, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt16Slot", 1, 3, 3, 2); __PYX_ERR(0, 671, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependInt16Slot") < 0)) __PYX_ERR(0, 671, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int16_t(values[1]); if (unlikely((__pyx_v_x == ((int16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_int16_t(values[2]); if (unlikely((__pyx_v_d == ((int16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependInt16Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 671, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt16Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_38PrependInt16Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_38PrependInt16Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int16_t __pyx_v_x, int16_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt16Slot", 0);

  /* "flattr/cflattr/builder.pyx":672
 * 
 *     def PrependInt16Slot(self, Py_ssize_t o, int16_t x, int16_t d):
 *         self.PrependSlot(&Fb_int16_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt32Slot(self, Py_ssize_t o, int32_t x, int32_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int16_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":671
 *         self.PrependSlot(&Fb_int8_t, o, x, d)
 * 
 *     def PrependInt16Slot(self, Py_ssize_t o, int16_t x, int16_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int16_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":674
 *         self.PrependSlot(&Fb_int16_t, o, x, d)
 * 
 *     def PrependInt32Slot(self, Py_ssize_t o, int32_t x, int32_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int32_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_41PrependInt32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_41PrependInt32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int32_t __pyx_v_x;
  int32_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt32Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt32Slot", 1, 3, 3, 1); __PYX_ERR(0, 674, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt32Slot", 1, 3, 3, 2); __PYX_ERR(0, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependInt32Slot") < 0)) __PYX_ERR(0, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 674, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int32_t(values[1]); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 674, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_int32_t(values[2]); if (unlikely((__pyx_v_d == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependInt32Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt32Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_40PrependInt32Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_40PrependInt32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int32_t __pyx_v_x, int32_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt32Slot", 0);

  /* "flattr/cflattr/builder.pyx":675
 * 
 *     def PrependInt32Slot(self, Py_ssize_t o, int32_t x, int32_t d):
 *         self.PrependSlot(&Fb_int32_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt64Slot(self, Py_ssize_t o, int64_t x, int64_t d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int32_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":674
 *         self.PrependSlot(&Fb_int16_t, o, x, d)
 * 
 *     def PrependInt32Slot(self, Py_ssize_t o, int32_t x, int32_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int32_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":677
 *         self.PrependSlot(&Fb_int32_t, o, x, d)
 * 
 *     def PrependInt64Slot(self, Py_ssize_t o, int64_t x, int64_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int64_t, o, x, d)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_43PrependInt64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_43PrependInt64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  int64_t __pyx_v_x;
  int64_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt64Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt64Slot", 1, 3, 3, 1); __PYX_ERR(0, 677, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependInt64Slot", 1, 3, 3, 2); __PYX_ERR(0, 677, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependInt64Slot") < 0)) __PYX_ERR(0, 677, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyInt_As_int64_t(values[2]); if (unlikely((__pyx_v_d == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependInt64Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 677, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt64Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_42PrependInt64Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_42PrependInt64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, int64_t __pyx_v_x, int64_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt64Slot", 0);

  /* "flattr/cflattr/builder.pyx":678
 * 
 *     def PrependInt64Slot(self, Py_ssize_t o, int64_t x, int64_t d):
 *         self.PrependSlot(&Fb_int64_t, o, x, d)             # <<<<<<<<<<<<<<
 * 
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependSlot(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int64_t), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* "flattr/cflattr/builder.pyx":677
 *         self.PrependSlot(&Fb_int32_t, o, x, d)
 * 
 *     def PrependInt64Slot(self, Py_ssize_t o, int64_t x, int64_t d):             # <<<<<<<<<<<<<<
 *         self.PrependSlot(&Fb_int64_t, o, x, d)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":680
 *         self.PrependSlot(&Fb_int64_t, o, x, d)
 * 
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_45PrependFloat32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_45PrependFloat32Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  float __pyx_v_x;
  float __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependFloat32Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependFloat32Slot", 1, 3, 3, 1); __PYX_ERR(0, 680, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependFloat32Slot", 1, 3, 3, 2); __PYX_ERR(0, 680, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependFloat32Slot") < 0)) __PYX_ERR(0, 680, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_x = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_d = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_d == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependFloat32Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 680, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat32Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_44PrependFloat32Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_44PrependFloat32Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, float __pyx_v_x, float __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependFloat32Slot", 0);

  /* "flattr/cflattr/builder.pyx":681
 * 
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 *             self.head = self.head - Fb_float32_t.bytewidth
 */
  __pyx_t_1 = ((__pyx_v_x != __pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":682
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):
 *         if x != d:
 *             self.Prep(Fb_float32_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *             self.head = self.head - Fb_float32_t.bytewidth
 *             writeFloat32(x, self.buffer, self.head)
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t.bytewidth, 0);

    /* "flattr/cflattr/builder.pyx":683
 *         if x != d:
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 *             self.head = self.head - Fb_float32_t.bytewidth             # <<<<<<<<<<<<<<
 *             writeFloat32(x, self.buffer, self.head)
 *             self.Slot(o)
 */
    __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t.bytewidth);

    /* "flattr/cflattr/builder.pyx":684
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 *             self.head = self.head - Fb_float32_t.bytewidth
 *             writeFloat32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *             self.Slot(o)
 * 
 */
    __pyx_f_6flattr_7cflattr_7builder_writeFloat32(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":685
 *             self.head = self.head - Fb_float32_t.bytewidth
 *             writeFloat32(x, self.buffer, self.head)
 *             self.Slot(o)             # <<<<<<<<<<<<<<
 * 
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Slot(__pyx_v_self, __pyx_v_o); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 685, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":681
 * 
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 *             self.head = self.head - Fb_float32_t.bytewidth
 */
  }

  /* "flattr/cflattr/builder.pyx":680
 *         self.PrependSlot(&Fb_int64_t, o, x, d)
 * 
 *     def PrependFloat32Slot(self, Py_ssize_t o, float x, float d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prep(Fb_float32_t.bytewidth, 0)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat32Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":687
 *             self.Slot(o)
 * 
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_47PrependFloat64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_47PrependFloat64Slot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  double __pyx_v_x;
  double __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependFloat64Slot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependFloat64Slot", 1, 3, 3, 1); __PYX_ERR(0, 687, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependFloat64Slot", 1, 3, 3, 2); __PYX_ERR(0, 687, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependFloat64Slot") < 0)) __PYX_ERR(0, 687, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L3_error)
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L3_error)
    __pyx_v_d = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_d == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependFloat64Slot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 687, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat64Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_46PrependFloat64Slot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_46PrependFloat64Slot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, double __pyx_v_x, double __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependFloat64Slot", 0);

  /* "flattr/cflattr/builder.pyx":688
 * 
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 *             self.head = self.head - Fb_float64_t.bytewidth
 */
  __pyx_t_1 = ((__pyx_v_x != __pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":689
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):
 *         if x != d:
 *             self.Prep(Fb_float64_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *             self.head = self.head - Fb_float64_t.bytewidth
 *             writeFloat64(x, self.buffer, self.head)
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t.bytewidth, 0);

    /* "flattr/cflattr/builder.pyx":690
 *         if x != d:
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 *             self.head = self.head - Fb_float64_t.bytewidth             # <<<<<<<<<<<<<<
 *             writeFloat64(x, self.buffer, self.head)
 *             self.Slot(o)
 */
    __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t.bytewidth);

    /* "flattr/cflattr/builder.pyx":691
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 *             self.head = self.head - Fb_float64_t.bytewidth
 *             writeFloat64(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *             self.Slot(o)
 * 
 */
    __pyx_f_6flattr_7cflattr_7builder_writeFloat64(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":692
 *             self.head = self.head - Fb_float64_t.bytewidth
 *             writeFloat64(x, self.buffer, self.head)
 *             self.Slot(o)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUOffsetTRelativeSlot(self, Py_ssize_t o, Py_ssize_t x, Py_ssize_t d):
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Slot(__pyx_v_self, __pyx_v_o); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 692, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":688
 * 
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 *             self.head = self.head - Fb_float64_t.bytewidth
 */
  }

  /* "flattr/cflattr/builder.pyx":687
 *             self.Slot(o)
 * 
 *     def PrependFloat64Slot(self, Py_ssize_t o, double x, double d):             # <<<<<<<<<<<<<<
 *         if x != d:
 *             self.Prep(Fb_float64_t.bytewidth, 0)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat64Slot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":694
 *             self.Slot(o)
 * 
 *     def PrependUOffsetTRelativeSlot(self, Py_ssize_t o, Py_ssize_t x, Py_ssize_t d):             # <<<<<<<<<<<<<<
 *         """
 *         PrependUOffsetTRelativeSlot prepends an UOffsetT onto the object at
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_49PrependUOffsetTRelativeSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_48PrependUOffsetTRelativeSlot[] = "\n        PrependUOffsetTRelativeSlot prepends an UOffsetT onto the object at\n        vtable slot `o`. If value `x` equals default `d`, then the slot will\n        be set to zero and no other data will be written.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_49PrependUOffsetTRelativeSlot(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_o;
  Py_ssize_t __pyx_v_x;
  Py_ssize_t __pyx_v_d;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUOffsetTRelativeSlot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_x,&__pyx_n_s_d,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_o)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUOffsetTRelativeSlot", 1, 3, 3, 1); __PYX_ERR(0, 694, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("PrependUOffsetTRelativeSlot", 1, 3, 3, 2); __PYX_ERR(0, 694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "PrependUOffsetTRelativeSlot") < 0)) __PYX_ERR(0, 694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_o = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_o == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_x == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_d = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_d == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("PrependUOffsetTRelativeSlot", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUOffsetTRelativeSlot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_48PrependUOffsetTRelativeSlot(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), __pyx_v_o, __pyx_v_x, __pyx_v_d);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_48PrependUOffsetTRelativeSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, Py_ssize_t __pyx_v_o, Py_ssize_t __pyx_v_x, Py_ssize_t __pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUOffsetTRelativeSlot", 0);

  /* "flattr/cflattr/builder.pyx":700
 *         be set to zero and no other data will be written.
 *         """
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.PrependUOffsetTRelative(x)
 *             self.Slot(o)
 */
  __pyx_t_1 = ((__pyx_v_x != __pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":701
 *         """
 *         if x != d:
 *             self.PrependUOffsetTRelative(x)             # <<<<<<<<<<<<<<
 *             self.Slot(o)
 * 
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->PrependUOffsetTRelative(__pyx_v_self, __pyx_v_x, 0);

    /* "flattr/cflattr/builder.pyx":702
 *         if x != d:
 *             self.PrependUOffsetTRelative(x)
 *             self.Slot(o)             # <<<<<<<<<<<<<<
 * 
 *     cdef PrependStructSlot(self, v, x, d):
 */
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Slot(__pyx_v_self, __pyx_v_o); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":700
 *         be set to zero and no other data will be written.
 *         """
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.PrependUOffsetTRelative(x)
 *             self.Slot(o)
 */
  }

  /* "flattr/cflattr/builder.pyx":694
 *             self.Slot(o)
 * 
 *     def PrependUOffsetTRelativeSlot(self, Py_ssize_t o, Py_ssize_t x, Py_ssize_t d):             # <<<<<<<<<<<<<<
 *         """
 *         PrependUOffsetTRelativeSlot prepends an UOffsetT onto the object at
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUOffsetTRelativeSlot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":704
 *             self.Slot(o)
 * 
 *     cdef PrependStructSlot(self, v, x, d):             # <<<<<<<<<<<<<<
 *         """
 *         PrependStructSlot prepends a struct onto the object at vtable slot `o`.
 */

static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependStructSlot(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_v, PyObject *__pyx_v_x, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependStructSlot", 0);

  /* "flattr/cflattr/builder.pyx":710
 *         In generated code, `d` is always 0.
 *         """
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.assertStructIsInline(x)
 *             self.Slot(v)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_x, __pyx_v_d, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "flattr/cflattr/builder.pyx":711
 *         """
 *         if x != d:
 *             self.assertStructIsInline(x)             # <<<<<<<<<<<<<<
 *             self.Slot(v)
 * 
 */
    __pyx_t_3 = __Pyx_PyIndex_AsSsize_t(__pyx_v_x); if (unlikely((__pyx_t_3 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L1_error)
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->assertStructIsInline(__pyx_v_self, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":712
 *         if x != d:
 *             self.assertStructIsInline(x)
 *             self.Slot(v)             # <<<<<<<<<<<<<<
 * 
 *     def PrependBool(self, x):
 */
    __pyx_t_3 = __Pyx_PyIndex_AsSsize_t(__pyx_v_v); if (unlikely((__pyx_t_3 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L1_error)
    ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Slot(__pyx_v_self, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L1_error)

    /* "flattr/cflattr/builder.pyx":710
 *         In generated code, `d` is always 0.
 *         """
 *         if x != d:             # <<<<<<<<<<<<<<
 *             self.assertStructIsInline(x)
 *             self.Slot(v)
 */
  }

  /* "flattr/cflattr/builder.pyx":704
 *             self.Slot(o)
 * 
 *     cdef PrependStructSlot(self, v, x, d):             # <<<<<<<<<<<<<<
 *         """
 *         PrependStructSlot prepends a struct onto the object at vtable slot `o`.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependStructSlot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":714
 *             self.Slot(v)
 * 
 *     def PrependBool(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `bool` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_51PrependBool(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_50PrependBool[] = "Prepend a `bool` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_51PrependBool(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependBool (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_50PrependBool(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_50PrependBool(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependBool", 0);

  /* "flattr/cflattr/builder.pyx":719
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_bool_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependByte(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 719, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_bool_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":714
 *             self.Slot(v)
 * 
 *     def PrependBool(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `bool` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependBool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":721
 *         self.Prepend(&Fb_bool_t, x)
 * 
 *     def PrependByte(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `byte` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_53PrependByte(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_52PrependByte[] = "Prepend a `byte` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_53PrependByte(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependByte (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_52PrependByte(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_52PrependByte(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependByte", 0);

  /* "flattr/cflattr/builder.pyx":726
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_uint8_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint8(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":721
 *         self.Prepend(&Fb_bool_t, x)
 * 
 *     def PrependByte(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `byte` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependByte", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":728
 *         self.Prepend(&Fb_uint8_t, x)
 * 
 *     def PrependUint8(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint8` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_55PrependUint8(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_54PrependUint8[] = "Prepend an `uint8` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_55PrependUint8(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint8 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_54PrependUint8(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_54PrependUint8(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUint8", 0);

  /* "flattr/cflattr/builder.pyx":733
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_uint8_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint16(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 733, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":728
 *         self.Prepend(&Fb_uint8_t, x)
 * 
 *     def PrependUint8(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint8` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":735
 *         self.Prepend(&Fb_uint8_t, x)
 * 
 *     def PrependUint16(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint16` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_57PrependUint16(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_56PrependUint16[] = "Prepend an `uint16` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_57PrependUint16(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint16 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_56PrependUint16(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_56PrependUint16(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUint16", 0);

  /* "flattr/cflattr/builder.pyx":740
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_uint16_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint32(self, uint32_t x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":735
 *         self.Prepend(&Fb_uint8_t, x)
 * 
 *     def PrependUint16(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint16` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":742
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 *     def PrependUint32(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint32` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_59PrependUint32(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_58PrependUint32[] = "Prepend an `uint32` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_59PrependUint32(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint32_t __pyx_v_x;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint32 (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 742, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_58PrependUint32(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_58PrependUint32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint32", 0);

  /* "flattr/cflattr/builder.pyx":747
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_uint32_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependUint64(self, x):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t), __pyx_v_x);

  /* "flattr/cflattr/builder.pyx":742
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 *     def PrependUint32(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint32` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":749
 *         self.Prepend(&Fb_uint32_t, x)
 * 
 *     def PrependUint64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint64` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_61PrependUint64(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_60PrependUint64[] = "Prepend an `uint64` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_61PrependUint64(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependUint64 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_60PrependUint64(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_60PrependUint64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependUint64", 0);

  /* "flattr/cflattr/builder.pyx":754
 *         Note: aligns and checks for space.
 *         """
 *         self.Prep(Fb_uint64_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         self.Place(x, &Fb_uint64_t)
 * 
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t.bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":755
 *         """
 *         self.Prep(Fb_uint64_t.bytewidth, 0)
 *         self.Place(x, &Fb_uint64_t)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt8(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Place(__pyx_v_self, __pyx_t_1, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t));

  /* "flattr/cflattr/builder.pyx":749
 *         self.Prepend(&Fb_uint32_t, x)
 * 
 *     def PrependUint64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `uint64` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependUint64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":757
 *         self.Place(x, &Fb_uint64_t)
 * 
 *     def PrependInt8(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int8` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_63PrependInt8(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_62PrependInt8[] = "Prepend an `int8` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_63PrependInt8(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt8 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_62PrependInt8(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_62PrependInt8(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependInt8", 0);

  /* "flattr/cflattr/builder.pyx":762
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_int8_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt16(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int8_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":757
 *         self.Place(x, &Fb_uint64_t)
 * 
 *     def PrependInt8(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int8` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":764
 *         self.Prepend(&Fb_int8_t, x)
 * 
 *     def PrependInt16(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int16` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_65PrependInt16(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_64PrependInt16[] = "Prepend an `int16` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_65PrependInt16(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt16 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_64PrependInt16(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_64PrependInt16(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependInt16", 0);

  /* "flattr/cflattr/builder.pyx":769
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_int16_t, x)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PrependInt32(self, int32_t x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 769, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int16_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":764
 *         self.Prepend(&Fb_int8_t, x)
 * 
 *     def PrependInt16(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int16` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":771
 *         self.Prepend(&Fb_int16_t, x)
 * 
 *     cpdef PrependInt32(self, int32_t x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int32` to the Builder buffer.
 * 
 */

static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_67PrependInt32(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependInt32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependInt32", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_PrependInt32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6flattr_7cflattr_7builder_7Builder_67PrependInt32)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int32_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flattr/cflattr/builder.pyx":776
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_int32_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependInt64(self, x):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int32_t), __pyx_v_x);

  /* "flattr/cflattr/builder.pyx":771
 *         self.Prepend(&Fb_int16_t, x)
 * 
 *     cpdef PrependInt32(self, int32_t x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int32` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_67PrependInt32(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_66PrependInt32[] = "Prepend an `int32` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_67PrependInt32(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int32_t __pyx_v_x;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt32 (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 771, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_66PrependInt32(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((int32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_66PrependInt32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependInt32", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependInt32(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":778
 *         self.Prepend(&Fb_int32_t, x)
 * 
 *     def PrependInt64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int64` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_69PrependInt64(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_68PrependInt64[] = "Prepend an `int64` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_69PrependInt64(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependInt64 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_68PrependInt64(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_68PrependInt64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependInt64", 0);

  /* "flattr/cflattr/builder.pyx":783
 *         Note: aligns and checks for space.
 *         """
 *         self.Prepend(&Fb_int64_t, x)             # <<<<<<<<<<<<<<
 * 
 *     def PrependFloat32(self, x):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(__pyx_v_x); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L1_error)
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_int64_t), __pyx_t_1);

  /* "flattr/cflattr/builder.pyx":778
 *         self.Prepend(&Fb_int32_t, x)
 * 
 *     def PrependInt64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend an `int64` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependInt64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":785
 *         self.Prepend(&Fb_int64_t, x)
 * 
 *     def PrependFloat32(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `float32` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_71PrependFloat32(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_70PrependFloat32[] = "Prepend a `float32` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_71PrependFloat32(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependFloat32 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_70PrependFloat32(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_70PrependFloat32(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  float __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependFloat32", 0);

  /* "flattr/cflattr/builder.pyx":790
 *         Note: aligns and checks for space.
 *         """
 *         self.Prep(Fb_float32_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         self.head = self.head - Fb_float32_t.bytewidth
 *         writeFloat32(x, self.buffer, self.head)
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t.bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":791
 *         """
 *         self.Prep(Fb_float32_t.bytewidth, 0)
 *         self.head = self.head - Fb_float32_t.bytewidth             # <<<<<<<<<<<<<<
 *         writeFloat32(x, self.buffer, self.head)
 * 
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":792
 *         self.Prep(Fb_float32_t.bytewidth, 0)
 *         self.head = self.head - Fb_float32_t.bytewidth
 *         writeFloat32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 * 
 *     def PrependFloat64(self, x):
 */
  __pyx_t_1 = __pyx_PyFloat_AsFloat(__pyx_v_x); if (unlikely((__pyx_t_1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L1_error)
  __pyx_f_6flattr_7cflattr_7builder_writeFloat32(__pyx_t_1, __pyx_v_self->buffer, __pyx_v_self->head, 0);

  /* "flattr/cflattr/builder.pyx":785
 *         self.Prepend(&Fb_int64_t, x)
 * 
 *     def PrependFloat32(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `float32` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":794
 *         writeFloat32(x, self.buffer, self.head)
 * 
 *     def PrependFloat64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `float64` to the Builder buffer.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_73PrependFloat64(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6flattr_7cflattr_7builder_7Builder_72PrependFloat64[] = "Prepend a `float64` to the Builder buffer.\n\n        Note: aligns and checks for space.\n        ";
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_73PrependFloat64(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependFloat64 (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_72PrependFloat64(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_72PrependFloat64(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PrependFloat64", 0);

  /* "flattr/cflattr/builder.pyx":799
 *         Note: aligns and checks for space.
 *         """
 *         self.Prep(Fb_float64_t.bytewidth, 0)             # <<<<<<<<<<<<<<
 *         self.head = self.head - Fb_float64_t.bytewidth
 *         writeFloat64(x, self.buffer, self.head)
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prep(__pyx_v_self, __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t.bytewidth, 0);

  /* "flattr/cflattr/builder.pyx":800
 *         """
 *         self.Prep(Fb_float64_t.bytewidth, 0)
 *         self.head = self.head - Fb_float64_t.bytewidth             # <<<<<<<<<<<<<<
 *         writeFloat64(x, self.buffer, self.head)
 * 
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":801
 *         self.Prep(Fb_float64_t.bytewidth, 0)
 *         self.head = self.head - Fb_float64_t.bytewidth
 *         writeFloat64(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PrependVOffsetT(self, uint16_t x):
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_x); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 801, __pyx_L1_error)
  __pyx_f_6flattr_7cflattr_7builder_writeFloat64(__pyx_t_1, __pyx_v_self->buffer, __pyx_v_self->head, 0);

  /* "flattr/cflattr/builder.pyx":794
 *         writeFloat32(x, self.buffer, self.head)
 * 
 *     def PrependFloat64(self, x):             # <<<<<<<<<<<<<<
 *         """Prepend a `float64` to the Builder buffer.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.PrependFloat64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flattr/cflattr/builder.pyx":803
 *         writeFloat64(x, self.buffer, self.head)
 * 
 *     cdef void PrependVOffsetT(self, uint16_t x):             # <<<<<<<<<<<<<<
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PrependVOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint16_t __pyx_v_x) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PrependVOffsetT", 0);

  /* "flattr/cflattr/builder.pyx":804
 * 
 *     cdef void PrependVOffsetT(self, uint16_t x):
 *         self.Prepend(&Fb_uint16_t, x)             # <<<<<<<<<<<<<<
 * 
 *     cdef void Place(self, uint64_t x, FlatbufferType *fb_type):
 */
  ((struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder *)__pyx_v_self->__pyx_vtab)->Prepend(__pyx_v_self, (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t), __pyx_v_x);

  /* "flattr/cflattr/builder.pyx":803
 *         writeFloat64(x, self.buffer, self.head)
 * 
 *     cdef void PrependVOffsetT(self, uint16_t x):             # <<<<<<<<<<<<<<
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":806
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 *     cdef void Place(self, uint64_t x, FlatbufferType *fb_type):             # <<<<<<<<<<<<<<
 *         """
 *         Place prepends a value specified by `flags` to the Builder,
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_Place(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint64_t __pyx_v_x, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *__pyx_v_fb_type) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("Place", 0);

  /* "flattr/cflattr/builder.pyx":811
 *         without checking for available space.
 *         """
 *         self.head = self.head - fb_type.bytewidth             # <<<<<<<<<<<<<<
 *         #Write(flags.packer_type, self.Bytes, self.head, x)
 * 
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_fb_type->bytewidth);

  /* "flattr/cflattr/builder.pyx":814
 *         #Write(flags.packer_type, self.Bytes, self.head, x)
 * 
 *         if fb_type == &Fb_bool_t:             # <<<<<<<<<<<<<<
 *             writeBool(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint8_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_bool_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":815
 * 
 *         if fb_type == &Fb_bool_t:
 *             writeBool(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_uint8_t:
 *             writeUint8(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeBool(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":814
 *         #Write(flags.packer_type, self.Bytes, self.head, x)
 * 
 *         if fb_type == &Fb_bool_t:             # <<<<<<<<<<<<<<
 *             writeBool(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint8_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":816
 *         if fb_type == &Fb_bool_t:
 *             writeBool(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint8_t:             # <<<<<<<<<<<<<<
 *             writeUint8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint16_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":817
 *             writeBool(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint8_t:
 *             writeUint8(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_uint16_t:
 *             writeUint16(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeUint8(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":816
 *         if fb_type == &Fb_bool_t:
 *             writeBool(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint8_t:             # <<<<<<<<<<<<<<
 *             writeUint8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint16_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":818
 *         elif fb_type == &Fb_uint8_t:
 *             writeUint8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint16_t:             # <<<<<<<<<<<<<<
 *             writeUint16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint32_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":819
 *             writeUint8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint16_t:
 *             writeUint16(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_uint32_t:
 *             writeUint32(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeUint16(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":818
 *         elif fb_type == &Fb_uint8_t:
 *             writeUint8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint16_t:             # <<<<<<<<<<<<<<
 *             writeUint16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint32_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":820
 *         elif fb_type == &Fb_uint16_t:
 *             writeUint16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint32_t:             # <<<<<<<<<<<<<<
 *             writeUint32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint64_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":821
 *             writeUint16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint32_t:
 *             writeUint32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_uint64_t:
 *             writeUint64(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeUint32(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":820
 *         elif fb_type == &Fb_uint16_t:
 *             writeUint16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint32_t:             # <<<<<<<<<<<<<<
 *             writeUint32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint64_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":822
 *         elif fb_type == &Fb_uint32_t:
 *             writeUint32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint64_t:             # <<<<<<<<<<<<<<
 *             writeUint64(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int8_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":823
 *             writeUint32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint64_t:
 *             writeUint64(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_int8_t:
 *             writeInt8(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeUint64(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":822
 *         elif fb_type == &Fb_uint32_t:
 *             writeUint32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_uint64_t:             # <<<<<<<<<<<<<<
 *             writeUint64(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int8_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":824
 *         elif fb_type == &Fb_uint64_t:
 *             writeUint64(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int8_t:             # <<<<<<<<<<<<<<
 *             writeInt8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int16_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_int8_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":825
 *             writeUint64(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int8_t:
 *             writeInt8(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_int16_t:
 *             writeInt16(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeInt8(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":824
 *         elif fb_type == &Fb_uint64_t:
 *             writeUint64(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int8_t:             # <<<<<<<<<<<<<<
 *             writeInt8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int16_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":826
 *         elif fb_type == &Fb_int8_t:
 *             writeInt8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int16_t:             # <<<<<<<<<<<<<<
 *             writeInt16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int32_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_int16_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":827
 *             writeInt8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int16_t:
 *             writeInt16(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_int32_t:
 *             writeInt32(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeInt16(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":826
 *         elif fb_type == &Fb_int8_t:
 *             writeInt8(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int16_t:             # <<<<<<<<<<<<<<
 *             writeInt16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int32_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":828
 *         elif fb_type == &Fb_int16_t:
 *             writeInt16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int32_t:             # <<<<<<<<<<<<<<
 *             writeInt32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int64_t:
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_int32_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":829
 *             writeInt16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int32_t:
 *             writeInt32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 *         elif fb_type == &Fb_int64_t:
 *             writeInt64(x, self.buffer, self.head)
 */
    __pyx_f_6flattr_7cflattr_7builder_writeInt32(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":828
 *         elif fb_type == &Fb_int16_t:
 *             writeInt16(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int32_t:             # <<<<<<<<<<<<<<
 *             writeInt32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int64_t:
 */
    goto __pyx_L3;
  }

  /* "flattr/cflattr/builder.pyx":830
 *         elif fb_type == &Fb_int32_t:
 *             writeInt32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int64_t:             # <<<<<<<<<<<<<<
 *             writeInt64(x, self.buffer, self.head)
 * 
 */
  __pyx_t_1 = ((__pyx_v_fb_type == (&__pyx_v_6flattr_7cflattr_7builder_Fb_int64_t)) != 0);
  if (__pyx_t_1) {

    /* "flattr/cflattr/builder.pyx":831
 *             writeInt32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int64_t:
 *             writeInt64(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PlaceVOffsetT(self, uint16_t x):
 */
    __pyx_f_6flattr_7cflattr_7builder_writeInt64(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

    /* "flattr/cflattr/builder.pyx":830
 *         elif fb_type == &Fb_int32_t:
 *             writeInt32(x, self.buffer, self.head)
 *         elif fb_type == &Fb_int64_t:             # <<<<<<<<<<<<<<
 *             writeInt64(x, self.buffer, self.head)
 * 
 */
  }
  __pyx_L3:;

  /* "flattr/cflattr/builder.pyx":806
 *         self.Prepend(&Fb_uint16_t, x)
 * 
 *     cdef void Place(self, uint64_t x, FlatbufferType *fb_type):             # <<<<<<<<<<<<<<
 *         """
 *         Place prepends a value specified by `flags` to the Builder,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":833
 *             writeInt64(x, self.buffer, self.head)
 * 
 *     cdef void PlaceVOffsetT(self, uint16_t x):             # <<<<<<<<<<<<<<
 *         """PlaceVOffsetT prepends a VOffsetT to the Builder, without checking
 *         for space.
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceVOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint16_t __pyx_v_x) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PlaceVOffsetT", 0);

  /* "flattr/cflattr/builder.pyx":837
 *         for space.
 *         """
 *         self.head = self.head - Fb_uint16_t.bytewidth             # <<<<<<<<<<<<<<
 *         writeUint16(x, self.buffer, self.head)
 * 
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":838
 *         """
 *         self.head = self.head - Fb_uint16_t.bytewidth
 *         writeUint16(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PlaceSOffsetT(self, int32_t x):
 */
  __pyx_f_6flattr_7cflattr_7builder_writeUint16(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

  /* "flattr/cflattr/builder.pyx":833
 *             writeInt64(x, self.buffer, self.head)
 * 
 *     cdef void PlaceVOffsetT(self, uint16_t x):             # <<<<<<<<<<<<<<
 *         """PlaceVOffsetT prepends a VOffsetT to the Builder, without checking
 *         for space.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":840
 *         writeUint16(x, self.buffer, self.head)
 * 
 *     cdef void PlaceSOffsetT(self, int32_t x):             # <<<<<<<<<<<<<<
 *         """PlaceSOffsetT prepends a SOffsetT to the Builder, without checking
 *         for space.
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceSOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, int32_t __pyx_v_x) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PlaceSOffsetT", 0);

  /* "flattr/cflattr/builder.pyx":844
 *         for space.
 *         """
 *         self.head = self.head - Fb_int32_t.bytewidth             # <<<<<<<<<<<<<<
 *         #Write(soffset, self.Bytes, self.head, x)
 *         writeInt32(x, self.buffer, self.head)
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":846
 *         self.head = self.head - Fb_int32_t.bytewidth
 *         #Write(soffset, self.Bytes, self.head, x)
 *         writeInt32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 * 
 *     cdef void PlaceUOffsetT(self, uint32_t x):
 */
  __pyx_f_6flattr_7cflattr_7builder_writeInt32(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

  /* "flattr/cflattr/builder.pyx":840
 *         writeUint16(x, self.buffer, self.head)
 * 
 *     cdef void PlaceSOffsetT(self, int32_t x):             # <<<<<<<<<<<<<<
 *         """PlaceSOffsetT prepends a SOffsetT to the Builder, without checking
 *         for space.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flattr/cflattr/builder.pyx":848
 *         writeInt32(x, self.buffer, self.head)
 * 
 *     cdef void PlaceUOffsetT(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         """PlaceUOffsetT prepends a UOffsetT to the Builder, without checking
 *         for space.
 */

static void __pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceUOffsetT(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, uint32_t __pyx_v_x) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PlaceUOffsetT", 0);

  /* "flattr/cflattr/builder.pyx":852
 *         for space.
 *         """
 *         self.head = self.head - Fb_uint32_t.bytewidth             # <<<<<<<<<<<<<<
 *         #Write(uoffset, self.Bytes, self.head, x)
 *         writeUint32(x, self.buffer, self.head)
 */
  __pyx_v_self->head = (__pyx_v_self->head - __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t.bytewidth);

  /* "flattr/cflattr/builder.pyx":854
 *         self.head = self.head - Fb_uint32_t.bytewidth
 *         #Write(uoffset, self.Bytes, self.head, x)
 *         writeUint32(x, self.buffer, self.head)             # <<<<<<<<<<<<<<
 */
  __pyx_f_6flattr_7cflattr_7builder_writeUint32(__pyx_v_x, __pyx_v_self->buffer, __pyx_v_self->head, 0);

  /* "flattr/cflattr/builder.pyx":848
 *         writeInt32(x, self.buffer, self.head)
 * 
 *     cdef void PlaceUOffsetT(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         """PlaceUOffsetT prepends a UOffsetT to the Builder, without checking
 *         for space.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_75__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_75__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_74__reduce_cython__(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_74__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_77__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6flattr_7cflattr_7builder_7Builder_77__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6flattr_7cflattr_7builder_7Builder_76__setstate_cython__(((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6flattr_7cflattr_7builder_7Builder_76__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6flattr_7cflattr_7builder_Builder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flattr.cflattr.builder.Builder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "number_types.pxd":13
 * 
 * 
 * cdef inline void enforce_number(uint64_t n, Flags flags) except *:             # <<<<<<<<<<<<<<
 *     return
 */

static CYTHON_INLINE void __pyx_f_6flattr_7cflattr_12number_types_enforce_number(CYTHON_UNUSED uint64_t __pyx_v_n, CYTHON_UNUSED struct __pyx_obj_6flattr_7cflattr_12number_types_Flags *__pyx_v_flags) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enforce_number", 0);

  /* "number_types.pxd":14
 * 
 * cdef inline void enforce_number(uint64_t n, Flags flags) except *:
 *     return             # <<<<<<<<<<<<<<
 */
  goto __pyx_L0;

  /* "number_types.pxd":13
 * 
 * 
 * cdef inline void enforce_number(uint64_t n, Flags flags) except *:             # <<<<<<<<<<<<<<
 *     return
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}
static struct __pyx_vtabstruct_6flattr_7cflattr_7builder_Builder __pyx_vtable_6flattr_7cflattr_7builder_Builder;

static PyObject *__pyx_tp_new_6flattr_7cflattr_7builder_Builder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6flattr_7cflattr_7builder_Builder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6flattr_7cflattr_7builder_Builder *)o);
  p->__pyx_vtab = __pyx_vtabptr_6flattr_7cflattr_7builder_Builder;
  return o;
}

static void __pyx_tp_dealloc_6flattr_7cflattr_7builder_Builder(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6flattr_7cflattr_7builder_7Builder_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6flattr_7cflattr_7builder_7Builder_Bytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6flattr_7cflattr_7builder_7Builder_5Bytes_1__get__(o);
}

static PyMethodDef __pyx_methods_6flattr_7cflattr_7builder_Builder[] = {
  {"Output", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_5Output, METH_NOARGS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_4Output},
  {"StartObject", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_7StartObject, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_6StartObject},
  {"EndObject", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_9EndObject, METH_NOARGS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_8EndObject},
  {"PrependUOffsetTRelative", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_11PrependUOffsetTRelative, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_10PrependUOffsetTRelative},
  {"StartVector", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_13StartVector, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_12StartVector},
  {"EndVector", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_15EndVector, METH_NOARGS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_14EndVector},
  {"CreateString", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_17CreateString, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_16CreateString},
  {"CreateByteVector", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_19CreateByteVector, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_18CreateByteVector},
  {"Finish", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_21Finish, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_20Finish},
  {"FinishSizePrefixed", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_23FinishSizePrefixed, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_22FinishSizePrefixed},
  {"PrependBoolSlot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_25PrependBoolSlot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependByteSlot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_27PrependByteSlot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependUint8Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_29PrependUint8Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependUint16Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_31PrependUint16Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependUint32Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_33PrependUint32Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependUint64Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_35PrependUint64Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependInt8Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_37PrependInt8Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependInt16Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_39PrependInt16Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependInt32Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_41PrependInt32Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependInt64Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_43PrependInt64Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependFloat32Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_45PrependFloat32Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependFloat64Slot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_47PrependFloat64Slot, METH_VARARGS|METH_KEYWORDS, 0},
  {"PrependUOffsetTRelativeSlot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7Builder_49PrependUOffsetTRelativeSlot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6flattr_7cflattr_7builder_7Builder_48PrependUOffsetTRelativeSlot},
  {"PrependBool", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_51PrependBool, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_50PrependBool},
  {"PrependByte", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_53PrependByte, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_52PrependByte},
  {"PrependUint8", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_55PrependUint8, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_54PrependUint8},
  {"PrependUint16", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_57PrependUint16, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_56PrependUint16},
  {"PrependUint32", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_59PrependUint32, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_58PrependUint32},
  {"PrependUint64", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_61PrependUint64, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_60PrependUint64},
  {"PrependInt8", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_63PrependInt8, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_62PrependInt8},
  {"PrependInt16", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_65PrependInt16, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_64PrependInt16},
  {"PrependInt32", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_67PrependInt32, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_66PrependInt32},
  {"PrependInt64", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_69PrependInt64, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_68PrependInt64},
  {"PrependFloat32", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_71PrependFloat32, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_70PrependFloat32},
  {"PrependFloat64", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_73PrependFloat64, METH_O, __pyx_doc_6flattr_7cflattr_7builder_7Builder_72PrependFloat64},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_75__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6flattr_7cflattr_7builder_7Builder_77__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6flattr_7cflattr_7builder_Builder[] = {
  {(char *)"Bytes", __pyx_getprop_6flattr_7cflattr_7builder_7Builder_Bytes, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6flattr_7cflattr_7builder_Builder = {
  PyVarObject_HEAD_INIT(0, 0)
  "flattr.cflattr.builder.Builder", /*tp_name*/
  sizeof(struct __pyx_obj_6flattr_7cflattr_7builder_Builder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6flattr_7cflattr_7builder_Builder, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  " A Builder is used to construct one or more FlatBuffers.\n\n    Typically, Builder objects will be used from code generated by the `flatc`\n    compiler.\n\n    A Builder constructs byte buffers in a last-first manner for simplicity and\n    performance during reading.\n\n    Internally, a Builder is a state machine for creating FlatBuffer objects.\n\n    It holds the following internal state:\n        - Bytes: an array of bytes.\n        - current_vtable: a list of integers.\n        - vtables: a list of vtable entries (i.e. a list of list of integers).\n\n    Attributes:\n      Bytes: The internal `bytearray` for the Builder.\n      finished: A boolean determining if the Builder has been finalized.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6flattr_7cflattr_7builder_Builder, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6flattr_7cflattr_7builder_Builder, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6flattr_7cflattr_7builder_7Builder_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6flattr_7cflattr_7builder_Builder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"writeBool", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_1writeBool, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeUint8", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_3writeUint8, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeUint16", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_5writeUint16, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeUint32", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_7writeUint32, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeUint64", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_9writeUint64, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeInt8", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_11writeInt8, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeInt16", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_13writeInt16, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeInt32", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_15writeInt32, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeInt64", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_17writeInt64, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeFloat64", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_19writeFloat64, METH_VARARGS|METH_KEYWORDS, 0},
  {"writeFloat32", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6flattr_7cflattr_7builder_21writeFloat32, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_builder(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_builder},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "builder",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Builder, __pyx_k_Builder, sizeof(__pyx_k_Builder), 0, 0, 1, 1},
  {&__pyx_n_s_BuilderNotFinishedError, __pyx_k_BuilderNotFinishedError, sizeof(__pyx_k_BuilderNotFinishedError), 0, 0, 1, 1},
  {&__pyx_n_s_BuilderSizeError, __pyx_k_BuilderSizeError, sizeof(__pyx_k_BuilderSizeError), 0, 0, 1, 1},
  {&__pyx_n_s_CreateByteVector, __pyx_k_CreateByteVector, sizeof(__pyx_k_CreateByteVector), 0, 0, 1, 1},
  {&__pyx_n_s_CreateString, __pyx_k_CreateString, sizeof(__pyx_k_CreateString), 0, 0, 1, 1},
  {&__pyx_n_s_EndVector, __pyx_k_EndVector, sizeof(__pyx_k_EndVector), 0, 0, 1, 1},
  {&__pyx_kp_s_Error_caused_by_an_Offset_arith, __pyx_k_Error_caused_by_an_Offset_arith, sizeof(__pyx_k_Error_caused_by_an_Offset_arith), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_caused_by_causing_a_Build, __pyx_k_Error_caused_by_causing_a_Build, sizeof(__pyx_k_Error_caused_by_causing_a_Build), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_caused_by_not_calling_Fin, __pyx_k_Error_caused_by_not_calling_Fin, sizeof(__pyx_k_Error_caused_by_not_calling_Fin), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_caused_by_using_a_Builder, __pyx_k_Error_caused_by_using_a_Builder, sizeof(__pyx_k_Error_caused_by_using_a_Builder), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_caused_by_using_a_Builder_2, __pyx_k_Error_caused_by_using_a_Builder_2, sizeof(__pyx_k_Error_caused_by_using_a_Builder_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_caused_by_using_a_Builder_3, __pyx_k_Error_caused_by_using_a_Builder_3, sizeof(__pyx_k_Error_caused_by_using_a_Builder_3), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IsNestedError, __pyx_k_IsNestedError, sizeof(__pyx_k_IsNestedError), 0, 0, 1, 1},
  {&__pyx_n_s_IsNotNestedError, __pyx_k_IsNotNestedError, sizeof(__pyx_k_IsNotNestedError), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_OffsetArithmeticError, __pyx_k_OffsetArithmeticError, sizeof(__pyx_k_OffsetArithmeticError), 0, 0, 1, 1},
  {&__pyx_n_s_PrependInt32, __pyx_k_PrependInt32, sizeof(__pyx_k_PrependInt32), 0, 0, 1, 1},
  {&__pyx_n_s_PrependUOffsetTRelative, __pyx_k_PrependUOffsetTRelative, sizeof(__pyx_k_PrependUOffsetTRelative), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_StructIsNotInlineError, __pyx_k_StructIsNotInlineError, sizeof(__pyx_k_StructIsNotInlineError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_alignment, __pyx_k_alignment, sizeof(__pyx_k_alignment), 0, 0, 1, 1},
  {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_bytewidth, __pyx_k_bytewidth, sizeof(__pyx_k_bytewidth), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_elemSize, __pyx_k_elemSize, sizeof(__pyx_k_elemSize), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
  {&__pyx_n_s_flatbuffers, __pyx_k_flatbuffers, sizeof(__pyx_k_flatbuffers), 0, 0, 1, 1},
  {&__pyx_kp_u_flatbuffers_Cannot_create_Builde, __pyx_k_flatbuffers_Cannot_create_Builde, sizeof(__pyx_k_flatbuffers_Cannot_create_Builde), 0, 1, 0, 0},
  {&__pyx_kp_u_flatbuffers_Offset_arithmetic_er, __pyx_k_flatbuffers_Offset_arithmetic_er, sizeof(__pyx_k_flatbuffers_Offset_arithmetic_er), 0, 1, 0, 0},
  {&__pyx_kp_u_flatbuffers_Tried_to_write_a_Str, __pyx_k_flatbuffers_Tried_to_write_a_Str, sizeof(__pyx_k_flatbuffers_Tried_to_write_a_Str), 0, 1, 0, 0},
  {&__pyx_kp_u_flatbuffers_cannot_grow_buffer_b, __pyx_k_flatbuffers_cannot_grow_buffer_b, sizeof(__pyx_k_flatbuffers_cannot_grow_buffer_b), 0, 1, 0, 0},
  {&__pyx_n_s_flattr_cflattr_builder, __pyx_k_flattr_cflattr_builder, sizeof(__pyx_k_flattr_cflattr_builder), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_head, __pyx_k_head, sizeof(__pyx_k_head), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_initialSize, __pyx_k_initialSize, sizeof(__pyx_k_initialSize), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_numElems, __pyx_k_numElems, sizeof(__pyx_k_numElems), 0, 0, 1, 1},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_n_s_pack_into, __pyx_k_pack_into, sizeof(__pyx_k_pack_into), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_kp_s_self_current_vtable_self_vtables, __pyx_k_self_current_vtable_self_vtables, sizeof(__pyx_k_self_current_vtable_self_vtables), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_u_strict, __pyx_k_strict, sizeof(__pyx_k_strict), 0, 1, 0, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 238, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 334, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 611, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_self_current_vtable_self_vtables); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.current_vtable,self.vtables cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_self_current_vtable_self_vtables); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "flattr/cflattr/builder.pyx":144
 * 
 * 
 * class OffsetArithmeticError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by an Offset arithmetic error. Probably caused by bad
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "flattr/cflattr/builder.pyx":153
 * 
 * 
 * class IsNotNestedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to write Object data when not inside
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "flattr/cflattr/builder.pyx":161
 * 
 * 
 * class IsNestedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to begin an Object when an Object is
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "flattr/cflattr/builder.pyx":169
 * 
 * 
 * class StructIsNotInlineError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to write a Struct at a location that
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "flattr/cflattr/builder.pyx":177
 * 
 * 
 * class BuilderSizeError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by causing a Builder to exceed the hardcoded limit of 2
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "flattr/cflattr/builder.pyx":184
 *     pass
 * 
 * class BuilderNotFinishedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by not calling `Finish` before calling `Output`.
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6flattr_7cflattr_7builder_Builder = &__pyx_vtable_6flattr_7cflattr_7builder_Builder;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.WriteVtable = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *))__pyx_f_6flattr_7cflattr_7builder_7Builder_WriteVtable;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.growByteBuffer = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *))__pyx_f_6flattr_7cflattr_7builder_7Builder_growByteBuffer;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Offset = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *))__pyx_f_6flattr_7cflattr_7builder_7Builder_Offset;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Pad = (PyObject *(*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_Pad;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Prep = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, Py_ssize_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_Prep;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependSOffsetTRelative = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSOffsetTRelative;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependUOffsetTRelative = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependUOffsetTRelative;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.EndVector = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int __pyx_skip_dispatch))__pyx_f_6flattr_7cflattr_7builder_7Builder_EndVector;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.CreateString = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6flattr_7cflattr_7builder_7Builder_CreateString *__pyx_optional_args))__pyx_f_6flattr_7cflattr_7builder_7Builder_CreateString;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.CreateByteVector = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6flattr_7cflattr_7builder_7Builder_CreateByteVector;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.assertNested = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *))__pyx_f_6flattr_7cflattr_7builder_7Builder_assertNested;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.assertNotNested = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *))__pyx_f_6flattr_7cflattr_7builder_7Builder_assertNotNested;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.assertStructIsInline = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_assertStructIsInline;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Slot = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_Slot;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.__pyx___Finish = (Py_ssize_t (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, Py_ssize_t, int))__pyx_f_6flattr_7cflattr_7builder_7Builder___Finish;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Prepend = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *, int64_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_Prepend;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependSlot = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *, Py_ssize_t, int64_t, int64_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependSlot;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependStructSlot = (PyObject *(*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, PyObject *, PyObject *, PyObject *))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependStructSlot;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependInt32 = (PyObject *(*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int32_t, int __pyx_skip_dispatch))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependInt32;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PrependVOffsetT = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint16_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_PrependVOffsetT;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.Place = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint64_t, struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType *))__pyx_f_6flattr_7cflattr_7builder_7Builder_Place;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PlaceVOffsetT = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint16_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceVOffsetT;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PlaceSOffsetT = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, int32_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceSOffsetT;
  __pyx_vtable_6flattr_7cflattr_7builder_Builder.PlaceUOffsetT = (void (*)(struct __pyx_obj_6flattr_7cflattr_7builder_Builder *, uint32_t))__pyx_f_6flattr_7cflattr_7builder_7Builder_PlaceUOffsetT;
  if (PyType_Ready(&__pyx_type_6flattr_7cflattr_7builder_Builder) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6flattr_7cflattr_7builder_Builder.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6flattr_7cflattr_7builder_Builder.tp_dictoffset && __pyx_type_6flattr_7cflattr_7builder_Builder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6flattr_7cflattr_7builder_Builder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_6flattr_7cflattr_7builder_Builder, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 195, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6flattr_7cflattr_7builder_7Builder___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6flattr_7cflattr_7builder_7Builder___init__.doc = __pyx_doc_6flattr_7cflattr_7builder_7Builder___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6flattr_7cflattr_7builder_7Builder___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_6flattr_7cflattr_7builder_Builder.tp_dict, __pyx_vtabptr_6flattr_7cflattr_7builder_Builder) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Builder, (PyObject *)&__pyx_type_6flattr_7cflattr_7builder_Builder) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6flattr_7cflattr_7builder_Builder) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_ptype_6flattr_7cflattr_7builder_Builder = &__pyx_type_6flattr_7cflattr_7builder_Builder;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flattr.cflattr.number_types"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6flattr_7cflattr_12number_types_Flags = __Pyx_ImportType(__pyx_t_1, "flattr.cflattr.number_types", "Flags", sizeof(struct __pyx_obj_6flattr_7cflattr_12number_types_Flags), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6flattr_7cflattr_12number_types_Flags) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initbuilder(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initbuilder(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_builder(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_builder(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_builder(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_t_6flattr_7cflattr_7builder_FlatbufferType __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'builder' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_builder(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("builder", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flattr__cflattr__builder) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flattr.cflattr.builder")) {
      if (unlikely(PyDict_SetItemString(modules, "flattr.cflattr.builder", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "flattr/cflattr/builder.pyx":19
 * from cpython.mem cimport PyMem_Malloc, PyMem_Free
 * 
 * import struct             # <<<<<<<<<<<<<<
 * 
 * from .number_types cimport Flags, enforce_number
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flattr/cflattr/builder.pyx":23
 * from .number_types cimport Flags, enforce_number
 * 
 * from flatbuffers import encode             # <<<<<<<<<<<<<<
 * 
 * cdef struct FlatbufferType:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_encode);
  __Pyx_GIVEREF(__pyx_n_s_encode);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_encode);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_flatbuffers, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_encode, __pyx_t_1) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":36
 * 
 * 
 * cdef FlatbufferType Fb_bool_t = FlatbufferType(1)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_uint8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)
 */
  __pyx_t_3.bytewidth = 1;
  __pyx_v_6flattr_7cflattr_7builder_Fb_bool_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":37
 * 
 * cdef FlatbufferType Fb_bool_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_uint8_t = FlatbufferType(1)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)
 */
  __pyx_t_3.bytewidth = 1;
  __pyx_v_6flattr_7cflattr_7builder_Fb_uint8_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":38
 * cdef FlatbufferType Fb_bool_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_uint8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)
 */
  __pyx_t_3.bytewidth = 2;
  __pyx_v_6flattr_7cflattr_7builder_Fb_uint16_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":39
 * cdef FlatbufferType Fb_uint8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_int8_t = FlatbufferType(1)
 */
  __pyx_t_3.bytewidth = 4;
  __pyx_v_6flattr_7cflattr_7builder_Fb_uint32_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":40
 * cdef FlatbufferType Fb_uint16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_int8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_int16_t = FlatbufferType(2)
 */
  __pyx_t_3.bytewidth = 8;
  __pyx_v_6flattr_7cflattr_7builder_Fb_uint64_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":41
 * cdef FlatbufferType Fb_uint32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_int8_t = FlatbufferType(1)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_int16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_int32_t = FlatbufferType(4)
 */
  __pyx_t_3.bytewidth = 1;
  __pyx_v_6flattr_7cflattr_7builder_Fb_int8_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":42
 * cdef FlatbufferType Fb_uint64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_int8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_int16_t = FlatbufferType(2)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_int32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_int64_t = FlatbufferType(8)
 */
  __pyx_t_3.bytewidth = 2;
  __pyx_v_6flattr_7cflattr_7builder_Fb_int16_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":43
 * cdef FlatbufferType Fb_int8_t = FlatbufferType(1)
 * cdef FlatbufferType Fb_int16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_int32_t = FlatbufferType(4)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_int64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_float32_t = FlatbufferType(4)
 */
  __pyx_t_3.bytewidth = 4;
  __pyx_v_6flattr_7cflattr_7builder_Fb_int32_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":44
 * cdef FlatbufferType Fb_int16_t = FlatbufferType(2)
 * cdef FlatbufferType Fb_int32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_int64_t = FlatbufferType(8)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_float32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_float64_t = FlatbufferType(8)
 */
  __pyx_t_3.bytewidth = 8;
  __pyx_v_6flattr_7cflattr_7builder_Fb_int64_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":45
 * cdef FlatbufferType Fb_int32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_int64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_float32_t = FlatbufferType(4)             # <<<<<<<<<<<<<<
 * cdef FlatbufferType Fb_float64_t = FlatbufferType(8)
 * 
 */
  __pyx_t_3.bytewidth = 4;
  __pyx_v_6flattr_7cflattr_7builder_Fb_float32_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":46
 * cdef FlatbufferType Fb_int64_t = FlatbufferType(8)
 * cdef FlatbufferType Fb_float32_t = FlatbufferType(4)
 * cdef FlatbufferType Fb_float64_t = FlatbufferType(8)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3.bytewidth = 8;
  __pyx_v_6flattr_7cflattr_7builder_Fb_float64_t = __pyx_t_3;

  /* "flattr/cflattr/builder.pyx":144
 * 
 * 
 * class OffsetArithmeticError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by an Offset arithmetic error. Probably caused by bad
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__3, __pyx_n_s_OffsetArithmeticError, __pyx_n_s_OffsetArithmeticError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_an_Offset_arith); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_OffsetArithmeticError, __pyx_tuple__3, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OffsetArithmeticError, __pyx_t_4) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":153
 * 
 * 
 * class IsNotNestedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to write Object data when not inside
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__4, __pyx_n_s_IsNotNestedError, __pyx_n_s_IsNotNestedError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_using_a_Builder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_IsNotNestedError, __pyx_tuple__4, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IsNotNestedError, __pyx_t_4) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":161
 * 
 * 
 * class IsNestedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to begin an Object when an Object is
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__5, __pyx_n_s_IsNestedError, __pyx_n_s_IsNestedError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_using_a_Builder_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_IsNestedError, __pyx_tuple__5, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IsNestedError, __pyx_t_4) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":169
 * 
 * 
 * class StructIsNotInlineError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by using a Builder to write a Struct at a location that
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__6, __pyx_n_s_StructIsNotInlineError, __pyx_n_s_StructIsNotInlineError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_using_a_Builder_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_StructIsNotInlineError, __pyx_tuple__6, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StructIsNotInlineError, __pyx_t_4) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":177
 * 
 * 
 * class BuilderSizeError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by causing a Builder to exceed the hardcoded limit of 2
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__7, __pyx_n_s_BuilderSizeError, __pyx_n_s_BuilderSizeError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_causing_a_Build); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_BuilderSizeError, __pyx_tuple__7, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BuilderSizeError, __pyx_t_4) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":184
 *     pass
 * 
 * class BuilderNotFinishedError(RuntimeError):             # <<<<<<<<<<<<<<
 *     """
 *     Error caused by not calling `Finish` before calling `Output`.
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__8, __pyx_n_s_BuilderNotFinishedError, __pyx_n_s_BuilderNotFinishedError, (PyObject *) NULL, __pyx_n_s_flattr_cflattr_builder, __pyx_kp_s_Error_caused_by_not_calling_Fin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_BuilderNotFinishedError, __pyx_tuple__8, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BuilderNotFinishedError, __pyx_t_4) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flattr/cflattr/builder.pyx":192
 * 
 * # VtableMetadataFields is the count of metadata fields in each vtable.
 * cdef unsigned int VtableMetadataFields = 2             # <<<<<<<<<<<<<<
 * cdef unsigned int MAX_BUFFER_SIZE = 2**31
 * 
 */
  __pyx_v_6flattr_7cflattr_7builder_VtableMetadataFields = 2;

  /* "flattr/cflattr/builder.pyx":193
 * # VtableMetadataFields is the count of metadata fields in each vtable.
 * cdef unsigned int VtableMetadataFields = 2
 * cdef unsigned int MAX_BUFFER_SIZE = 2**31             # <<<<<<<<<<<<<<
 * 
 * cdef class Builder(object):
 */
  __pyx_v_6flattr_7cflattr_7builder_MAX_BUFFER_SIZE = 0x80000000;

  /* "flattr/cflattr/builder.pyx":1
 * #cython: language_level=3,             # <<<<<<<<<<<<<<
 * # Copyright 2014 Google Inc. All rights reserved.
 * #
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "number_types.pxd":13
 * 
 * 
 * cdef inline void enforce_number(uint64_t n, Flags flags) except *:             # <<<<<<<<<<<<<<
 *     return
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init flattr.cflattr.builder", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flattr.cflattr.builder");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint16_t) -1;
        }
    } else {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyInt_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint64_t) -1;
        }
    } else {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 2 * PyLong_SHIFT) {
                            return (int8_t) (((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 3 * PyLong_SHIFT) {
                            return (int8_t) (((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 4 * PyLong_SHIFT) {
                            return (int8_t) (((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) ((((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) ((((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) ((((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int8_t) -1;
        }
    } else {
        int8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int8_t) -1;
        val = __Pyx_PyInt_As_int8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int8_t");
    return (int8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int8_t");
    return (int8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int16_t neg_one = (int16_t) -1, const_zero = (int16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 2 * PyLong_SHIFT) {
                            return (int16_t) (((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 3 * PyLong_SHIFT) {
                            return (int16_t) (((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 4 * PyLong_SHIFT) {
                            return (int16_t) (((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) ((((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) ((((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) ((((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int16_t) -1;
        }
    } else {
        int16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int16_t) -1;
        val = __Pyx_PyInt_As_int16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int16_t");
    return (int16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int16_t");
    return (int16_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 2 * PyLong_SHIFT) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 3 * PyLong_SHIFT) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 4 * PyLong_SHIFT) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 2 * PyLong_SHIFT) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 3 * PyLong_SHIFT) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 4 * PyLong_SHIFT) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
